"""SQLAlchemy ORM model integration for Moltres."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Sequence, Type, Union

if TYPE_CHECKING:
    from sqlalchemy.orm import DeclarativeBase
    from sqlalchemy import MetaData, Table
    from sqlalchemy.sql.type_api import TypeEngine

from .schema import (
    ColumnDef,
    TableSchema,
    UniqueConstraint,
    CheckConstraint,
    ForeignKeyConstraint,
)


def is_sqlalchemy_model(obj: Any) -> bool:
    """Detect if an object is a SQLAlchemy ORM model class.

    Args:
        obj: Object to check

    Returns:
        True if obj is a SQLAlchemy model class, False otherwise
    """
    try:
        from sqlalchemy.orm import DeclarativeMeta
        from sqlalchemy.inspection import inspect

        # Check if it's a class
        if not isinstance(obj, type):
            return False

        # Check if it has __table__ attribute (SQLAlchemy ORM models have this)
        if hasattr(obj, "__table__"):
            # Try to inspect it - if it's a model, inspect will work
            try:
                inspect(obj)
                return True
            except Exception:
                return False

        # Also check for DeclarativeMeta (older SQLAlchemy versions)
        if isinstance(obj, DeclarativeMeta):
            return True

        return False
    except ImportError:
        return False


def get_model_table_name(model_class: Type) -> str:
    """Extract table name from a SQLAlchemy model class.

    Args:
        model_class: SQLAlchemy model class

    Returns:
        Table name

    Raises:
        ValueError: If model doesn't have a table name
    """
    if hasattr(model_class, "__tablename__"):
        tablename = getattr(model_class, "__tablename__")
        if isinstance(tablename, str):
            return tablename
    if hasattr(model_class, "__table__"):
        # Fallback to __table__.name
        table = getattr(model_class, "__table__")
        if table is not None and hasattr(table, "name"):
            return str(table.name)
    # Fallback to class name (lowercased)
    return model_class.__name__.lower()


def sqlalchemy_type_to_moltres_type(sa_type: "TypeEngine") -> str:
    """Convert SQLAlchemy type to Moltres type name.

    Args:
        sa_type: SQLAlchemy TypeEngine instance

    Returns:
        Moltres type name string
    """
    from sqlalchemy import types as sa_types

    # Get the actual type class (handle wrapped types)
    type_class = type(sa_type)
    type_name = type_class.__name__

    # Handle common SQLAlchemy types
    if type_class == sa_types.Integer:
        return "INTEGER"
    elif type_class == sa_types.BigInteger:
        return "BIGINT"
    elif type_class == sa_types.SmallInteger:
        return "SMALLINT"
    elif type_class == sa_types.String:
        # Check if it has a length
        if hasattr(sa_type, "length") and sa_type.length is not None:
            return f"VARCHAR({sa_type.length})"
        return "TEXT"
    elif type_class == sa_types.Text:
        return "TEXT"
    elif type_class == sa_types.CHAR:
        if hasattr(sa_type, "length") and sa_type.length is not None:
            return f"CHAR({sa_type.length})"
        return "CHAR(1)"
    elif type_class == sa_types.Boolean:
        return "BOOLEAN"
    elif type_class == sa_types.REAL:
        return "REAL"
    elif type_class == sa_types.Float:
        return "FLOAT"
    elif type_class == sa_types.Double:
        return "DOUBLE"
    elif type_class == sa_types.Numeric:
        # Extract precision and scale
        precision = getattr(sa_type, "precision", None) or 10
        scale = getattr(sa_type, "scale", None) or 0
        return f"DECIMAL({precision},{scale})"
    elif type_class == sa_types.Date:
        return "DATE"
    elif type_class == sa_types.Time:
        return "TIME"
    elif type_class == sa_types.DateTime:
        return "DATETIME"
    elif type_class == sa_types.TIMESTAMP:
        return "TIMESTAMP"
    elif type_class == sa_types.Interval:
        return "INTERVAL"
    # Handle dialect-specific types
    elif hasattr(sa_types, "JSON") and type_class == sa_types.JSON:
        return "JSON"
    elif type_name == "UUID":
        return "UUID"
    elif type_name == "JSONB":
        return "JSONB"
    else:
        # Try to get dialect-specific types
        try:
            from sqlalchemy.dialects.postgresql import JSONB, UUID as PG_UUID

            if type_class == JSONB:
                return "JSONB"
            elif type_class == PG_UUID:
                return "UUID"
        except ImportError:
            pass

        # Fallback to TEXT for unknown types
        return "TEXT"


def moltres_type_to_sqlalchemy_type(
    type_name: str,
    nullable: bool = True,
    precision: Optional[int] = None,
    scale: Optional[int] = None,
) -> "TypeEngine":
    """Convert Moltres type name to SQLAlchemy TypeEngine.

    Args:
        type_name: Moltres type name
        nullable: Whether the column is nullable
        precision: Optional precision for DECIMAL/NUMERIC
        scale: Optional scale for DECIMAL/NUMERIC

    Returns:
        SQLAlchemy TypeEngine instance
    """
    from sqlalchemy import types as sa_types

    type_name_upper = type_name.upper()

    # Handle types with parameters (e.g., VARCHAR(100), DECIMAL(10,2))
    if "(" in type_name_upper:
        base_type = type_name_upper.split("(")[0]
        params = type_name_upper.split("(")[1].rstrip(")").split(",")
        if base_type == "VARCHAR":
            length = int(params[0].strip()) if params else None
            return sa_types.String(length=length)
        elif base_type == "CHAR":
            length = int(params[0].strip()) if params else 1
            return sa_types.CHAR(length=length)
        elif base_type in ("DECIMAL", "NUMERIC"):
            prec = int(params[0].strip()) if params else (precision or 10)
            sc = int(params[1].strip()) if len(params) > 1 else (scale or 0)
            return sa_types.Numeric(precision=prec, scale=sc)
        else:
            # Fallback to String for unknown parameterized types
            return sa_types.String()

    # Handle simple types
    if type_name_upper == "INTEGER":
        return sa_types.Integer()
    elif type_name_upper == "BIGINT":
        return sa_types.BigInteger()
    elif type_name_upper == "SMALLINT":
        return sa_types.SmallInteger()
    elif type_name_upper == "TEXT":
        return sa_types.Text()
    elif type_name_upper == "VARCHAR":
        return sa_types.String()
    elif type_name_upper == "CHAR":
        return sa_types.CHAR(length=1)
    elif type_name_upper == "BOOLEAN":
        return sa_types.Boolean()
    elif type_name_upper == "REAL":
        return sa_types.REAL()
    elif type_name_upper == "FLOAT":
        return sa_types.Float()
    elif type_name_upper == "DOUBLE":
        return sa_types.Double()
    elif type_name_upper in ("DECIMAL", "NUMERIC"):
        prec = precision or 10
        sc = scale or 0
        return sa_types.Numeric(precision=prec, scale=scale)
    elif type_name_upper == "DATE":
        return sa_types.Date()
    elif type_name_upper == "TIME":
        return sa_types.Time()
    elif type_name_upper == "DATETIME":
        return sa_types.DateTime()
    elif type_name_upper == "TIMESTAMP":
        return sa_types.TIMESTAMP()
    elif type_name_upper == "INTERVAL":
        return sa_types.Interval()
    elif type_name_upper == "UUID":
        try:
            from sqlalchemy.dialects.postgresql import UUID

            return UUID()
        except ImportError:
            return sa_types.CHAR(length=36)
    elif type_name_upper == "JSON":
        try:
            from sqlalchemy.dialects.postgresql import JSON

            return JSON()
        except ImportError:
            return sa_types.JSON()
    elif type_name_upper == "JSONB":
        try:
            from sqlalchemy.dialects.postgresql import JSONB

            return JSONB()
        except ImportError:
            return sa_types.JSON()
    else:
        # Fallback to String for unknown types
        return sa_types.String()


def extract_foreign_keys(model_class: Type) -> list[ForeignKeyConstraint]:
    """Extract foreign key constraints from a SQLAlchemy model.

    Args:
        model_class: SQLAlchemy model class

    Returns:
        List of ForeignKeyConstraint objects
    """
    from sqlalchemy import inspect as sa_inspect

    fk_constraints: list[ForeignKeyConstraint] = []

    try:
        mapper = sa_inspect(model_class)
        table = mapper.tables[0] if mapper.tables else None

        if table is None:
            return fk_constraints

        # Extract foreign keys from columns
        for column in table.columns:
            for fk in column.foreign_keys:
                # Get referenced table and column
                ref_table = fk.column.table.name
                ref_column = fk.column.name

                # Get ondelete and onupdate actions
                on_delete = fk.ondelete if hasattr(fk, "ondelete") else None
                on_update = fk.onupdate if hasattr(fk, "onupdate") else None

                # Convert ondelete/onupdate to strings
                on_delete_str = str(on_delete).upper() if on_delete else None
                on_update_str = str(on_update).upper() if on_update else None

                fk_constraints.append(
                    ForeignKeyConstraint(
                        name=None,  # SQLAlchemy doesn't always expose FK constraint names
                        columns=column.name,
                        references_table=ref_table,
                        references_columns=ref_column,
                        on_delete=on_delete_str,
                        on_update=on_update_str,
                    )
                )
    except Exception:
        # If inspection fails, return empty list
        pass

    return fk_constraints


def model_to_schema(model_class: Type["DeclarativeBase"]) -> TableSchema:
    """Convert a SQLAlchemy ORM model to a Moltres TableSchema.

    Args:
        model_class: SQLAlchemy model class

    Returns:
        TableSchema object

    Raises:
        ValueError: If model_class is not a valid SQLAlchemy model
    """
    from sqlalchemy import inspect as sa_inspect

    if not is_sqlalchemy_model(model_class):
        raise ValueError(f"{model_class} is not a valid SQLAlchemy model")

    # Get table name
    table_name = get_model_table_name(model_class)

    # Inspect the model
    mapper = sa_inspect(model_class)
    table = mapper.tables[0] if mapper.tables else None

    if table is None:
        raise ValueError(f"Model {model_class} does not have a table definition")

    # Convert columns
    columns: list[ColumnDef] = []
    for column in table.columns:
        # Get type name
        type_name = sqlalchemy_type_to_moltres_type(column.type)

        # Extract precision and scale for DECIMAL types
        precision = None
        scale = None
        if "DECIMAL" in type_name.upper() or "NUMERIC" in type_name.upper():
            if hasattr(column.type, "precision"):
                precision = column.type.precision
            if hasattr(column.type, "scale"):
                scale = column.type.scale

        # Get default value
        default = None
        if column.server_default is not None:
            # Extract default value from server_default
            if hasattr(column.server_default, "arg"):
                default = column.server_default.arg
            else:
                default = str(column.server_default)

        columns.append(
            ColumnDef(
                name=column.name,
                type_name=type_name,
                nullable=column.nullable,
                default=default,
                primary_key=column.primary_key,
                precision=precision,
                scale=scale,
            )
        )

    # Extract constraints
    constraints: list[Union[UniqueConstraint, CheckConstraint, ForeignKeyConstraint]] = []

    # Extract unique constraints from __table_args__
    if hasattr(model_class, "__table_args__"):
        table_args = model_class.__table_args__
        if isinstance(table_args, tuple):
            for arg in table_args:
                if isinstance(arg, type) and hasattr(arg, "__name__"):
                    # Skip type annotations
                    continue
                if hasattr(arg, "columns"):
                    # UniqueConstraint
                    from sqlalchemy import UniqueConstraint as SAUniqueConstraint

                    if isinstance(arg, SAUniqueConstraint):
                        cols = list(arg.columns.keys()) if hasattr(arg.columns, "keys") else []
                        if not cols:
                            # Try to get column names from the constraint
                            cols = (
                                [col.name for col in arg.columns]
                                if hasattr(arg.columns, "__iter__")
                                else []
                            )
                        # Handle SQLAlchemy's _NoneName type for unnamed constraints
                        unique_constraint_name: Optional[str] = None
                        if arg.name is not None:
                            name_str = str(arg.name)
                            if name_str and not name_str.startswith("_"):
                                unique_constraint_name = name_str
                        constraints.append(
                            UniqueConstraint(name=unique_constraint_name, columns=cols)
                        )
                elif hasattr(arg, "sqltext") or hasattr(arg, "sqltext"):
                    # CheckConstraint
                    from sqlalchemy import CheckConstraint as SACheckConstraint

                    if isinstance(arg, SACheckConstraint):
                        expr = str(arg.sqltext) if hasattr(arg, "sqltext") else ""
                        # Handle SQLAlchemy's _NoneName type for unnamed constraints
                        check_constraint_name: Optional[str] = None
                        if arg.name is not None:
                            name_str = str(arg.name)
                            if name_str and not name_str.startswith("_"):
                                check_constraint_name = name_str
                        constraints.append(
                            CheckConstraint(name=check_constraint_name, expression=expr)
                        )

    # Extract foreign keys
    fk_constraints = extract_foreign_keys(model_class)
    constraints.extend(fk_constraints)

    return TableSchema(
        name=table_name,
        columns=columns,
        if_not_exists=True,
        temporary=False,
        constraints=constraints,
    )


def schema_to_table(schema: TableSchema, metadata: "MetaData") -> "Table":
    """Convert a Moltres TableSchema to a SQLAlchemy Table object.

    Args:
        schema: TableSchema to convert
        metadata: SQLAlchemy MetaData object to attach the table to

    Returns:
        SQLAlchemy Table object
    """
    from sqlalchemy import Table as SATable

    # Convert columns
    sa_columns = []
    for col_def in schema.columns:
        sa_type = moltres_type_to_sqlalchemy_type(
            col_def.type_name,
            nullable=col_def.nullable,
            precision=col_def.precision,
            scale=col_def.scale,
        )

        from sqlalchemy import Column

        column_kwargs: dict[str, Any] = {
            "nullable": col_def.nullable,
            "primary_key": col_def.primary_key,
        }

        if col_def.default is not None:
            from sqlalchemy import text

            column_kwargs["server_default"] = text(str(col_def.default))

        sa_columns.append(Column(col_def.name, sa_type, **column_kwargs))

    # Convert constraints
    sa_constraints: list[Any] = []
    for constraint in schema.constraints:
        if isinstance(constraint, UniqueConstraint):
            from sqlalchemy import UniqueConstraint as SAUniqueConstraint

            cols = (
                list(constraint.columns)
                if isinstance(constraint.columns, Sequence)
                and not isinstance(constraint.columns, str)
                else [constraint.columns]
            )
            sa_constraints.append(SAUniqueConstraint(*cols, name=constraint.name))
        elif isinstance(constraint, CheckConstraint):
            from sqlalchemy import CheckConstraint as SACheckConstraint
            from sqlalchemy import text

            sa_constraints.append(
                SACheckConstraint(text(constraint.expression), name=constraint.name)
            )
        elif isinstance(constraint, ForeignKeyConstraint):
            from sqlalchemy import ForeignKey

            cols = (
                list(constraint.columns)
                if isinstance(constraint.columns, Sequence)
                and not isinstance(constraint.columns, str)
                else [constraint.columns]
            )
            ref_cols = (
                list(constraint.references_columns)
                if isinstance(constraint.references_columns, Sequence)
                and not isinstance(constraint.references_columns, str)
                else [constraint.references_columns]
            )
            # Create ForeignKey objects for each column
            refs = [f"{constraint.references_table}.{ref_col}" for ref_col in ref_cols]
            for i, col in enumerate(cols):
                fk = ForeignKey(
                    refs[i] if i < len(refs) else refs[0],
                    ondelete=constraint.on_delete,
                    onupdate=constraint.on_update,
                )
                # Find the column and add FK to it
                for sa_col in sa_columns:
                    if sa_col.name == col:
                        sa_col.foreign_keys.add(fk)
                        break

    # Create table
    table = SATable(
        schema.name,
        metadata,
        *sa_columns,
        *sa_constraints,
    )

    return table



"""Dataset readers."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any

from ..table.table import Database


def read_table(
    db: Database, table_name: str, columns: Iterable[str] | None = None
) -> list[dict[str, Any]] | Any:
    """Read a table from the database.

    Args:
        db: Database connection
        table_name: Name of the table to read
        columns: Optional list of column names to select. If None, selects all columns.

    Returns:
        Query results in the format specified by the database's fetch_format
        (list of dicts, pandas DataFrame, or polars DataFrame)
    """
    handle = db.table(table_name)
    df = handle.select()
    if columns:
        df = df.select(*columns)
    return df.collect()


def read_sql(
    db: Database, sql: str, params: dict[str, object] | None = None
) -> list[dict[str, Any]] | Any:
    """Execute a raw SQL query and return results.

    Args:
        db: Database connection
        sql: SQL query string
        params: Optional parameter dictionary for parameterized queries

    Returns:
        Query results in the format specified by the database's fetch_format
        (list of dicts, pandas DataFrame, or polars DataFrame)

    Raises:
        ExecutionError: If SQL execution fails
    """
    return db.execute_sql(sql, params=params).rows



"""Records class for file read operations."""

from __future__ import annotations

import logging
from collections.abc import AsyncIterator, Iterator, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, overload

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..table.async_table import AsyncDatabase, AsyncTableHandle
    from ..table.schema import ColumnDef
    from ..table.table import Database, TableHandle


def _is_pandas_dataframe(obj: Any) -> bool:
    """Check if object is a pandas DataFrame without importing pandas unnecessarily."""
    module = getattr(type(obj), "__module__", "")
    if "pandas" not in module:
        return False
    try:
        import pandas as pd
    except ImportError:
        return False
    return isinstance(obj, pd.DataFrame)


def _is_polars_dataframe(obj: Any) -> bool:
    """Check if object is a polars DataFrame."""
    module = getattr(type(obj), "__module__", "")
    if "polars" not in module:
        return False
    try:
        import polars as pl
    except ImportError:
        return False
    return isinstance(obj, pl.DataFrame)


def _is_polars_lazyframe(obj: Any) -> bool:
    """Check if object is a polars LazyFrame."""
    module = getattr(type(obj), "__module__", "")
    if "polars" not in module:
        return False
    try:
        import polars as pl
    except ImportError:
        return False
    return isinstance(obj, pl.LazyFrame)


def _convert_pandas_dtype_to_sql_type(dtype: Any) -> str:
    """Convert pandas dtype to SQL type name.

    Args:
        dtype: pandas dtype object

    Returns:
        SQL type name string
    """
    dtype_str = str(dtype)

    # Handle nullable integer types
    if dtype_str.startswith("Int"):
        return "INTEGER"

    # Handle standard types
    if dtype_str in ("int64", "int32", "int16", "int8", "int"):
        return "INTEGER"
    if dtype_str in ("float64", "float32", "float"):
        return "REAL"
    if dtype_str in ("bool", "boolean"):
        return "INTEGER"  # SQLite uses INTEGER for booleans
    if dtype_str.startswith("datetime"):
        return "TIMESTAMP"
    if dtype_str.startswith("date"):
        return "DATE"
    if dtype_str.startswith("timedelta"):
        return "TEXT"  # No standard SQL type for timedelta
    if dtype_str in ("object", "string", "str"):
        return "TEXT"

    # Default to TEXT for unknown types
    return "TEXT"


def _convert_polars_type_to_sql_type(polars_type: Any) -> str:
    """Convert polars type to SQL type name.

    Args:
        polars_type: polars DataType object or string representation

    Returns:
        SQL type name string
    """
    type_str = str(polars_type).lower()

    # Handle polars types
    if "int" in type_str:
        return "INTEGER"
    if "float" in type_str or "f64" in type_str or "f32" in type_str:
        return "REAL"
    if "bool" in type_str:
        return "INTEGER"  # SQLite uses INTEGER for booleans
    if "datetime" in type_str or "timestamp" in type_str:
        return "TIMESTAMP"
    if "date" in type_str:
        return "DATE"
    if "duration" in type_str or "timedelta" in type_str:
        return "TEXT"  # No standard SQL type for duration
    if "str" in type_str or "string" in type_str or "utf8" in type_str:
        return "TEXT"
    if "null" in type_str:
        return "TEXT"  # Nullable type, default to TEXT

    # Default to TEXT for unknown types
    return "TEXT"


def _extract_schema_from_pandas_dataframe(df: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from pandas DataFrame.

    Args:
        df: pandas DataFrame

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        for col_name, dtype in df.dtypes.items():
            sql_type = _convert_pandas_dtype_to_sql_type(dtype)
            # Check if column has any nulls - convert numpy boolean to Python boolean
            nullable = bool(df[col_name].isna().any())
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from pandas DataFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during pandas DataFrame schema extraction: %s", e, exc_info=True
        )
        return None


def _extract_schema_from_polars_dataframe(df: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from polars DataFrame.

    Args:
        df: polars DataFrame

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        schema = df.schema
        for col_name, polars_type in schema.items():
            sql_type = _convert_polars_type_to_sql_type(polars_type)
            # Check if type is nullable - convert to Python boolean
            nullable = bool(
                polars_type.is_nullable() if hasattr(polars_type, "is_nullable") else True
            )
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from polars DataFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during polars DataFrame schema extraction: %s", e, exc_info=True
        )
        return None


def _extract_schema_from_polars_lazyframe(lf: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from polars LazyFrame.

    Args:
        lf: polars LazyFrame

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        # Use collect_schema() to avoid performance warning
        schema = lf.collect_schema() if hasattr(lf, "collect_schema") else lf.schema
        for col_name, polars_type in schema.items():
            sql_type = _convert_polars_type_to_sql_type(polars_type)
            # Check if type is nullable - convert to Python boolean
            nullable = bool(
                polars_type.is_nullable() if hasattr(polars_type, "is_nullable") else True
            )
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from polars LazyFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during polars LazyFrame schema extraction: %s", e, exc_info=True
        )
        return None


def _convert_dataframe_to_rows(df: Any) -> List[dict[str, object]]:
    """Convert pandas/polars DataFrame or polars LazyFrame to list of dictionaries.

    Args:
        df: pandas DataFrame, polars DataFrame, or polars LazyFrame

    Returns:
        List of row dictionaries

    Raises:
        ValueError: If DataFrame type is not supported
    """
    if _is_pandas_dataframe(df):
        # Convert pandas DataFrame to list of dicts
        return df.to_dict("records")  # type: ignore[no-any-return]
    elif _is_polars_dataframe(df):
        # Convert polars DataFrame to list of dicts
        return df.to_dicts()  # type: ignore[no-any-return]
    elif _is_polars_lazyframe(df):
        # Materialize LazyFrame first, then convert
        materialized = df.collect()
        return materialized.to_dicts()  # type: ignore[no-any-return]
    else:
        raise ValueError(f"Unsupported DataFrame type: {type(df)}")


def _dataframe_to_records(df: Any, database: Optional["Database"] = None) -> "Records":
    """Convert pandas/polars DataFrame or polars LazyFrame to Records with lazy conversion.

    Args:
        df: pandas DataFrame, polars DataFrame, or polars LazyFrame
        database: Optional database reference

    Returns:
        Records object with lazy DataFrame conversion
    """
    # Extract schema if possible
    schema = None
    if _is_pandas_dataframe(df):
        schema = _extract_schema_from_pandas_dataframe(df)
    elif _is_polars_dataframe(df):
        schema = _extract_schema_from_polars_dataframe(df)
    elif _is_polars_lazyframe(df):
        schema = _extract_schema_from_polars_lazyframe(df)

    return Records(
        _data=None,
        _generator=None,
        _dataframe=df,
        _schema=schema,
        _database=database,
    )


@dataclass
class Records(Sequence[Mapping[str, object]]):
    """Container for file data that can be materialized or streaming.

    Records is NOT a DataFrame - it does not support SQL operations.
    It is designed for file reads and can be used with SQL insert operations.

    Attributes:
        _data: Materialized list of row dictionaries (for small files)
        _generator: Callable that returns an iterator of row chunks (for large files)
        _dataframe: Optional pandas/polars DataFrame or polars LazyFrame for lazy conversion
        _schema: Optional schema information
        _database: Optional database reference for insert operations
    """

    _data: Optional[List[dict[str, object]]] = None
    _generator: Optional[Callable[[], Iterator[List[dict[str, object]]]]] = None
    _dataframe: Optional[Any] = None  # pandas DataFrame, polars DataFrame, or polars LazyFrame
    _schema: Optional[Sequence["ColumnDef"]] = None
    _database: Optional["Database"] = None

    @classmethod
    def from_list(
        cls, data: List[dict[str, object]], database: Optional["Database"] = None
    ) -> "Records":
        """Create Records from a list of dictionaries.

        This is the recommended way to create Records from Python data.

        Args:
            data: List of row dictionaries
            database: Optional database reference for insert operations

        Returns:
            Records object

        Example:
            >>> records = Records.from_list(
            ...     [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}],
            ...     database=db
            ... )
            >>> records.insert_into("users")
        """
        return cls(_data=data, _database=database)

    @classmethod
    def from_dicts(
        cls, *dicts: dict[str, object], database: Optional["Database"] = None
    ) -> "Records":
        """Create Records from multiple dictionary arguments.

        Convenience method for creating Records from individual row dictionaries.

        Args:
            *dicts: Individual row dictionaries
            database: Optional database reference for insert operations

        Returns:
            Records object

        Example:
            >>> records = Records.from_dicts(
            ...     {"id": 1, "name": "Alice"},
            ...     {"id": 2, "name": "Bob"},
            ...     database=db
            ... )
            >>> records.insert_into("users")
        """
        return cls(_data=list(dicts), _database=database)

    @classmethod
    def from_dataframe(cls, df: Any, database: Optional["Database"] = None) -> "Records":
        """Create Records from pandas/polars DataFrame or polars LazyFrame.

        Args:
            df: pandas DataFrame, polars DataFrame, or polars LazyFrame
            database: Optional database reference for insert operations

        Returns:
            Records object with lazy DataFrame conversion

        Example:
            >>> import pandas as pd
            >>> df = pd.DataFrame([{"id": 1, "name": "Alice"}])
            >>> records = Records.from_dataframe(df, database=db)
            >>> records.insert_into("users")
        """
        return _dataframe_to_records(df, database=database)

    def __iter__(self) -> Iterator[dict[str, object]]:
        """Make Records directly iterable."""
        if self._data is not None:
            # Materialized mode - iterate over data
            for row in self._data:
                yield row
        elif self._generator is not None:
            # Streaming mode - iterate over generator chunks
            for chunk in self._generator():
                for row in chunk:
                    yield row
        elif self._dataframe is not None:
            # DataFrame mode - convert and iterate
            rows = _convert_dataframe_to_rows(self._dataframe)
            for row in rows:
                yield row
        # Empty records - nothing to yield

    def __len__(self) -> int:
        """Return the number of rows (materializes if needed)."""
        if self._data is not None:
            return len(self._data)
        elif self._generator is not None:
            # Materialize to get length
            count = 0
            for chunk in self._generator():
                count += len(chunk)
            return count
        elif self._dataframe is not None:
            # DataFrame mode - get length from DataFrame
            if _is_pandas_dataframe(self._dataframe):
                return len(self._dataframe)
            elif _is_polars_dataframe(self._dataframe):
                return len(self._dataframe)
            elif _is_polars_lazyframe(self._dataframe):
                # For LazyFrame, we need to materialize to get length
                # This is expensive, but necessary for __len__
                materialized = self._dataframe.collect()
                return len(materialized)
        return 0

    @overload
    def __getitem__(self, index: int) -> Mapping[str, object]: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Mapping[str, object]]: ...

    def __getitem__(
        self, index: int | slice
    ) -> Mapping[str, object] | Sequence[Mapping[str, object]]:
        """Get a row by index or slice (materializes if needed)."""
        if isinstance(index, slice):
            # For slices, materialize and return a list
            rows = self.rows()
            return rows[index]
        if self._data is not None:
            return self._data[index]
        elif self._generator is not None:
            # Materialize to get item
            rows = self.rows()
            return rows[index]
        elif self._dataframe is not None:
            # DataFrame mode - materialize to get item
            rows = self.rows()
            return rows[index]
        else:
            raise IndexError(
                "Cannot access Records: Records is empty. Use .rows() to check if data exists."
            )

    def rows(self) -> List[dict[str, object]]:
        """Return materialized list of all rows.

        Returns:
            List of row dictionaries
        """
        if self._data is not None:
            return self._data.copy()
        elif self._generator is not None:
            # Materialize from generator
            all_rows: List[dict[str, object]] = []
            for chunk in self._generator():
                all_rows.extend(chunk)
            return all_rows
        elif self._dataframe is not None:
            # DataFrame mode - convert to rows and cache in _data
            rows = _convert_dataframe_to_rows(self._dataframe)
            # Cache the converted data for future use
            self._data = rows
            self._dataframe = None  # Clear DataFrame reference after conversion
            return rows
        else:
            return []

    def iter(self) -> Iterator[dict[str, object]]:
        """Return an iterator over rows.

        Returns:
            Iterator of row dictionaries
        """
        return iter(self)

    @property
    def schema(self) -> Optional[Sequence["ColumnDef"]]:
        """Get the schema for these records."""
        return self._schema

    def select(self, *columns: str) -> "Records":
        """Select specific columns from records (in-memory operation).

        Args:
            *columns: Column names to select. Must be strings.

        Returns:
            New Records instance with only the selected columns

        Raises:
            ValueError: If no columns provided or column doesn't exist
            RuntimeError: If Records is empty

        Example:
            >>> records = Records(_data=[{"id": 1, "name": "Alice", "age": 30}], _database=db)
            >>> selected = records.select("id", "name")
            >>> list(selected)
            [{"id": 1, "name": "Alice"}]
        """
        if not columns:
            raise ValueError("select() requires at least one column name")

        rows = self.rows()
        if not rows:
            raise RuntimeError("Cannot select columns from empty Records")

        # Get all available columns from first row
        available_columns = set(rows[0].keys())

        # Validate all requested columns exist
        missing_columns = [col for col in columns if col not in available_columns]
        if missing_columns:
            available_str = ", ".join(sorted(available_columns))
            raise ValueError(
                f"Column(s) not found: {', '.join(missing_columns)}. "
                f"Available columns: {available_str}"
            )

        # Filter rows to only include selected columns
        filtered_rows = [{col: row[col] for col in columns} for row in rows]

        # Filter schema if available
        filtered_schema = None
        if self._schema is not None:
            schema_dict = {col.name: col for col in self._schema}
            filtered_schema = [schema_dict[col] for col in columns if col in schema_dict]

        return Records(
            _data=filtered_rows,
            _generator=None,
            _schema=filtered_schema,
            _database=self._database,
        )

    def rename(
        self, columns: Union[Dict[str, str], str], new_name: Optional[str] = None
    ) -> "Records":
        """Rename columns in records (in-memory operation).

        Args:
            columns: Either a dict mapping old_name -> new_name, or a single column name (if new_name provided)
            new_name: New name for the column (required if columns is a string)

        Returns:
            New Records instance with renamed columns

        Raises:
            ValueError: If column doesn't exist or new name conflicts with existing column
            RuntimeError: If Records is empty

        Example:
            >>> records = Records(_data=[{"id": 1, "name": "Alice"}], _database=db)
            >>> renamed = records.rename({"id": "user_id", "name": "user_name"})
            >>> list(renamed)
            [{"user_id": 1, "user_name": "Alice"}]

            >>> renamed = records.rename("id", "user_id")
            >>> list(renamed)
            [{"user_id": 1, "name": "Alice"}]
        """
        rows = self.rows()
        if not rows:
            raise RuntimeError("Cannot rename columns in empty Records")

        # Normalize to dict format
        if isinstance(columns, str):
            if new_name is None:
                raise ValueError("new_name is required when columns is a string")
            rename_map: Dict[str, str] = {columns: new_name}
        else:
            rename_map = columns

        if not rename_map:
            raise ValueError("rename() requires at least one column to rename")

        # Get all available columns from first row
        available_columns = set(rows[0].keys())

        # Validate all old columns exist
        missing_columns = [
            old_col for old_col in rename_map.keys() if old_col not in available_columns
        ]
        if missing_columns:
            available_str = ", ".join(sorted(available_columns))
            raise ValueError(
                f"Column(s) not found: {', '.join(missing_columns)}. "
                f"Available columns: {available_str}"
            )

        # Check for name conflicts (new name conflicts with existing column that's not being renamed)
        new_names = set(rename_map.values())
        conflicting = new_names & (available_columns - set(rename_map.keys()))
        if conflicting:
            raise ValueError(
                f"New column name(s) conflict with existing columns: {', '.join(conflicting)}"
            )

        # Rename columns in rows
        renamed_rows = []
        for row in rows:
            new_row = {}
            for key, value in row.items():
                if key in rename_map:
                    new_row[rename_map[key]] = value
                else:
                    new_row[key] = value
            renamed_rows.append(new_row)

        # Update schema if available
        updated_schema = None
        if self._schema is not None:
            from ..table.schema import ColumnDef

            updated_schema = []
            for col_def in self._schema:
                if col_def.name in rename_map:
                    updated_schema.append(
                        ColumnDef(
                            name=rename_map[col_def.name],
                            type_name=col_def.type_name,
                            nullable=col_def.nullable,
                        )
                    )
                else:
                    updated_schema.append(col_def)

        return Records(
            _data=renamed_rows,
            _generator=None,
            _schema=updated_schema,
            _database=self._database,
        )

    def head(self, n: int = 5) -> List[dict[str, object]]:
        """Return first n rows as list.

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of first n row dictionaries

        Raises:
            ValueError: If n is negative
        """
        if n < 0:
            raise ValueError(f"n must be non-negative, got {n}")
        rows = self.rows()
        return rows[:n]

    def tail(self, n: int = 5) -> List[dict[str, object]]:
        """Return last n rows as list.

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of last n row dictionaries

        Raises:
            ValueError: If n is negative
        """
        if n < 0:
            raise ValueError(f"n must be non-negative, got {n}")
        rows = self.rows()
        return rows[-n:]

    def first(self) -> Optional[dict[str, object]]:
        """Return first row or None if empty.

        Returns:
            First row dictionary or None if Records is empty
        """
        rows = self.rows()
        return rows[0] if rows else None

    def last(self) -> Optional[dict[str, object]]:
        """Return last row or None if empty.

        Returns:
            Last row dictionary or None if Records is empty
        """
        rows = self.rows()
        return rows[-1] if rows else None

    def insert_into(self, table: Union[str, "TableHandle"]) -> int:
        """Insert records into a table.

        Args:
            table: Table name (str) or TableHandle

        Returns:
            Number of rows inserted

        Raises:
            RuntimeError: If no database is attached

        Note:
            For DataFrame-based operations, consider creating a DataFrame from the data
            and using df.write.insertInto() instead.
        """
        if self._database is None:
            raise RuntimeError(
                "Cannot insert Records without an attached Database. "
                "For DataFrame-based operations, consider creating a DataFrame from the data "
                "and using df.write.insertInto() instead."
            )

        from ..table.mutations import insert_rows

        if isinstance(table, str):
            table_handle = self._database.table(table)
        else:
            table_handle = table

        table_handle_strict: "TableHandle" = table_handle
        transaction = self._database.connection_manager.active_transaction

        # Stream chunked data without materializing whenever possible
        if self._generator is not None:
            total_inserted = 0
            chunk_iter = self._generator()
            for chunk in chunk_iter:
                if not chunk:
                    continue
                total_inserted += insert_rows(table_handle_strict, chunk, transaction=transaction)
            return total_inserted

        rows = self.rows()
        if not rows:
            return 0
        return insert_rows(table_handle_strict, rows, transaction=transaction)


@dataclass
class AsyncRecords:
    """Async container for file data that can be materialized or streaming.

    AsyncRecords is NOT an AsyncDataFrame - it does not support SQL operations.
    It is designed for file reads and can be used with SQL insert operations.

    Attributes:
        _data: Materialized list of row dictionaries (for small files)
        _generator: Async callable that returns an async iterator of row chunks (for large files)
        _schema: Optional schema information
        _database: Optional database reference for insert operations
    """

    _data: Optional[List[dict[str, object]]] = None
    _generator: Optional[Callable[[], AsyncIterator[List[dict[str, object]]]]] = None
    _schema: Optional[Sequence["ColumnDef"]] = None
    _database: Optional["AsyncDatabase"] = None

    async def __aiter__(self) -> AsyncIterator[dict[str, object]]:
        """Make AsyncRecords directly async iterable."""
        if self._data is not None:
            # Materialized mode - iterate over data
            for row in self._data:
                yield row
        elif self._generator is not None:
            # Streaming mode - iterate over generator chunks
            async for chunk in self._generator():
                for row in chunk:
                    yield row
        # Empty records - nothing to yield

    async def rows(self) -> List[dict[str, object]]:
        """Return materialized list of all rows.

        Returns:
            List of row dictionaries
        """
        if self._data is not None:
            return self._data.copy()
        elif self._generator is not None:
            # Materialize from generator
            all_rows: List[dict[str, object]] = []
            async for chunk in self._generator():
                all_rows.extend(chunk)
            return all_rows
        else:
            return []

    async def iter(self) -> AsyncIterator[dict[str, object]]:
        """Return an async iterator over rows.

        Returns:
            AsyncIterator of row dictionaries
        """
        async for row in self:
            yield row

    @property
    def schema(self) -> Optional[Sequence["ColumnDef"]]:
        """Get the schema for these records."""
        return self._schema

    async def select(self, *columns: str) -> "AsyncRecords":
        """Select specific columns from records (in-memory operation).

        Args:
            *columns: Column names to select. Must be strings.

        Returns:
            New AsyncRecords instance with only the selected columns

        Raises:
            ValueError: If no columns provided or column doesn't exist
            RuntimeError: If AsyncRecords is empty

        Example:
            >>> records = AsyncRecords(_data=[{"id": 1, "name": "Alice", "age": 30}], _database=db)
            >>> selected = await records.select("id", "name")
            >>> async for row in selected:
            ...     print(row)
            {"id": 1, "name": "Alice"}
        """
        if not columns:
            raise ValueError("select() requires at least one column name")

        rows = await self.rows()
        if not rows:
            raise RuntimeError("Cannot select columns from empty AsyncRecords")

        # Get all available columns from first row
        available_columns = set(rows[0].keys())

        # Validate all requested columns exist
        missing_columns = [col for col in columns if col not in available_columns]
        if missing_columns:
            available_str = ", ".join(sorted(available_columns))
            raise ValueError(
                f"Column(s) not found: {', '.join(missing_columns)}. "
                f"Available columns: {available_str}"
            )

        # Filter rows to only include selected columns
        filtered_rows = [{col: row[col] for col in columns} for row in rows]

        # Filter schema if available
        filtered_schema = None
        if self._schema is not None:
            schema_dict = {col.name: col for col in self._schema}
            filtered_schema = [schema_dict[col] for col in columns if col in schema_dict]

        return AsyncRecords(
            _data=filtered_rows,
            _generator=None,
            _schema=filtered_schema,
            _database=self._database,
        )

    async def rename(
        self, columns: Union[Dict[str, str], str], new_name: Optional[str] = None
    ) -> "AsyncRecords":
        """Rename columns in records (in-memory operation).

        Args:
            columns: Either a dict mapping old_name -> new_name, or a single column name (if new_name provided)
            new_name: New name for the column (required if columns is a string)

        Returns:
            New AsyncRecords instance with renamed columns

        Raises:
            ValueError: If column doesn't exist or new name conflicts with existing column
            RuntimeError: If AsyncRecords is empty

        Example:
            >>> records = AsyncRecords(_data=[{"id": 1, "name": "Alice"}], _database=db)
            >>> renamed = await records.rename({"id": "user_id", "name": "user_name"})
            >>> async for row in renamed:
            ...     print(row)
            {"user_id": 1, "user_name": "Alice"}
        """
        rows = await self.rows()
        if not rows:
            raise RuntimeError("Cannot rename columns in empty AsyncRecords")

        # Normalize to dict format
        if isinstance(columns, str):
            if new_name is None:
                raise ValueError("new_name is required when columns is a string")
            rename_map: Dict[str, str] = {columns: new_name}
        else:
            rename_map = columns

        if not rename_map:
            raise ValueError("rename() requires at least one column to rename")

        # Get all available columns from first row
        available_columns = set(rows[0].keys())

        # Validate all old columns exist
        missing_columns = [
            old_col for old_col in rename_map.keys() if old_col not in available_columns
        ]
        if missing_columns:
            available_str = ", ".join(sorted(available_columns))
            raise ValueError(
                f"Column(s) not found: {', '.join(missing_columns)}. "
                f"Available columns: {available_str}"
            )

        # Check for name conflicts (new name conflicts with existing column that's not being renamed)
        new_names = set(rename_map.values())
        conflicting = new_names & (available_columns - set(rename_map.keys()))
        if conflicting:
            raise ValueError(
                f"New column name(s) conflict with existing columns: {', '.join(conflicting)}"
            )

        # Rename columns in rows
        renamed_rows = []
        for row in rows:
            new_row = {}
            for key, value in row.items():
                if key in rename_map:
                    new_row[rename_map[key]] = value
                else:
                    new_row[key] = value
            renamed_rows.append(new_row)

        # Update schema if available
        updated_schema = None
        if self._schema is not None:
            from ..table.schema import ColumnDef

            updated_schema = []
            for col_def in self._schema:
                if col_def.name in rename_map:
                    updated_schema.append(
                        ColumnDef(
                            name=rename_map[col_def.name],
                            type_name=col_def.type_name,
                            nullable=col_def.nullable,
                        )
                    )
                else:
                    updated_schema.append(col_def)

        return AsyncRecords(
            _data=renamed_rows,
            _generator=None,
            _schema=updated_schema,
            _database=self._database,
        )

    async def head(self, n: int = 5) -> List[dict[str, object]]:
        """Return first n rows as list.

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of first n row dictionaries

        Raises:
            ValueError: If n is negative
        """
        if n < 0:
            raise ValueError(f"n must be non-negative, got {n}")
        rows = await self.rows()
        return rows[:n]

    async def tail(self, n: int = 5) -> List[dict[str, object]]:
        """Return last n rows as list.

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of last n row dictionaries

        Raises:
            ValueError: If n is negative
        """
        if n < 0:
            raise ValueError(f"n must be non-negative, got {n}")
        rows = await self.rows()
        return rows[-n:]

    async def first(self) -> Optional[dict[str, object]]:
        """Return first row or None if empty.

        Returns:
            First row dictionary or None if AsyncRecords is empty
        """
        rows = await self.rows()
        return rows[0] if rows else None

    async def last(self) -> Optional[dict[str, object]]:
        """Return last row or None if empty.

        Returns:
            Last row dictionary or None if AsyncRecords is empty
        """
        rows = await self.rows()
        return rows[-1] if rows else None

    async def insert_into(self, table: Union[str, "AsyncTableHandle"]) -> int:
        """Insert records into a table.

        Args:
            table: Table name (str) or AsyncTableHandle

        Returns:
            Number of rows inserted

        Raises:
            RuntimeError: If no database is attached

        Note:
            For DataFrame-based operations, consider creating a DataFrame from the data
            and using df.write.insertInto() instead.
        """
        if self._database is None:
            raise RuntimeError(
                "Cannot insert AsyncRecords without an attached AsyncDatabase. "
                "For DataFrame-based operations, consider creating an AsyncDataFrame from the data "
                "and using df.write.insertInto() instead."
            )

        from ..table.async_mutations import insert_rows_async

        if isinstance(table, str):
            table_handle = await self._database.table(table)
        else:
            table_handle = table

        transaction = self._database.connection_manager.active_transaction

        if self._generator is not None:
            total_inserted = 0
            chunk_iter = self._generator()
            async for chunk in chunk_iter:
                if not chunk:
                    continue
                total_inserted += await insert_rows_async(
                    table_handle, chunk, transaction=transaction
                )
            return total_inserted

        rows = await self.rows()
        if not rows:
            return 0
        return await insert_rows_async(table_handle, rows, transaction=transaction)


@dataclass
class LazyRecords(Sequence[Mapping[str, object]]):
    """Lazy wrapper for Records that materializes on-demand.

    LazyRecords wraps a read operation and delays materialization until needed.
    It can be materialized explicitly with .collect() or automatically when:
    - Sequence operations are used (__len__, __getitem__, __iter__)
    - insert_into() is called
    - Used as argument to DataFrame operations

    Attributes:
        _read_func: Callable that returns Records when called (the read operation)
        _database: Database reference
        _schema: Optional schema information
        _options: Read options
        _materialized: Cached materialized Records (None until materialized)
    """

    _read_func: Callable[[], Records]
    _database: Optional["Database"]
    _schema: Optional[Sequence["ColumnDef"]] = None
    _options: Optional[dict[str, object]] = None
    _materialized: Optional[Records] = None

    def collect(self) -> Records:
        """Explicitly materialize and return Records.

        Returns:
            Materialized Records object
        """
        if self._materialized is None:
            self._materialized = self._read_func()
        return self._materialized

    def __iter__(self) -> Iterator[dict[str, object]]:
        """Make LazyRecords iterable (auto-materializes)."""
        return iter(self.collect())

    def __len__(self) -> int:
        """Return the number of rows (auto-materializes)."""
        return len(self.collect())

    @overload
    def __getitem__(self, index: int) -> Mapping[str, object]: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Mapping[str, object]]: ...

    def __getitem__(
        self, index: int | slice
    ) -> Mapping[str, object] | Sequence[Mapping[str, object]]:
        """Get a row by index or slice (auto-materializes)."""
        return self.collect()[index]

    def rows(self) -> List[dict[str, object]]:
        """Return materialized list of all rows (auto-materializes).

        Returns:
            List of row dictionaries
        """
        return self.collect().rows()

    def iter(self) -> Iterator[dict[str, object]]:
        """Return an iterator over rows (auto-materializes).

        Returns:
            Iterator of row dictionaries
        """
        return self.collect().iter()

    @property
    def schema(self) -> Optional[Sequence["ColumnDef"]]:
        """Get the schema for these records.

        Returns:
            Schema if available, None otherwise
        """
        # Try to get schema without materializing if possible
        if self._schema is not None:
            return self._schema
        # Otherwise materialize to get schema from Records
        return self.collect().schema

    def select(self, *columns: str) -> "Records":
        """Select specific columns from records (auto-materializes).

        Args:
            *columns: Column names to select. Must be strings.

        Returns:
            New Records with selected columns (materialized)

        Example:
            >>> lazy_records = LazyRecords(_read_func=lambda: Records(_data=[{"id": 1, "name": "Alice"}]))
            >>> selected = lazy_records.select("id")
            >>> list(selected)
            [{"id": 1}]
        """
        return self.collect().select(*columns)

    def rename(
        self, columns: Union[Dict[str, str], str], new_name: Optional[str] = None
    ) -> "Records":
        """Rename columns in records (auto-materializes).

        Args:
            columns: Either a dict mapping old_name -> new_name, or a single column name
            new_name: New name for the column (required if columns is a string)

        Returns:
            New Records with renamed columns (materialized)

        Example:
            >>> lazy_records = LazyRecords(_read_func=lambda: Records(_data=[{"id": 1}]))
            >>> renamed = lazy_records.rename("id", "user_id")
            >>> list(renamed)
            [{"user_id": 1}]
        """
        return self.collect().rename(columns, new_name)

    def head(self, n: int = 5) -> List[dict[str, object]]:
        """Return first n rows as list (auto-materializes).

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of first n row dictionaries
        """
        return self.collect().head(n)

    def tail(self, n: int = 5) -> List[dict[str, object]]:
        """Return last n rows as list (auto-materializes).

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of last n row dictionaries
        """
        return self.collect().tail(n)

    def first(self) -> Optional[dict[str, object]]:
        """Return first row or None if empty (auto-materializes).

        Returns:
            First row dictionary or None if LazyRecords is empty
        """
        return self.collect().first()

    def last(self) -> Optional[dict[str, object]]:
        """Return last row or None if empty (auto-materializes).

        Returns:
            Last row dictionary or None if LazyRecords is empty
        """
        return self.collect().last()

    def insert_into(self, table: Union[str, "TableHandle"]) -> int:
        """Insert records into a table (auto-materializes).

        Args:
            table: Table name (str) or TableHandle

        Returns:
            Number of rows inserted

        Raises:
            RuntimeError: If no database is attached
        """
        return self.collect().insert_into(table)


@dataclass
class AsyncLazyRecords:
    """Async lazy wrapper for AsyncRecords that materializes on-demand.

    AsyncLazyRecords wraps an async read operation and delays materialization until needed.
    It can be materialized explicitly with await .collect() or automatically when:
    - Async iteration is used (__aiter__)
    - insert_into() is called
    - Used as argument to async DataFrame operations

    Attributes:
        _read_func: Async callable (coroutine) that returns AsyncRecords when awaited
        _database: AsyncDatabase reference
        _schema: Optional schema information
        _options: Read options
        _materialized: Cached materialized AsyncRecords (None until materialized)
    """

    _read_func: Callable[[], Any]  # Returns a coroutine that returns AsyncRecords
    _database: Optional["AsyncDatabase"]
    _schema: Optional[Sequence["ColumnDef"]] = None
    _options: Optional[dict[str, object]] = None
    _materialized: Optional[AsyncRecords] = None

    async def collect(self) -> AsyncRecords:
        """Explicitly materialize and return AsyncRecords.

        Returns:
            Materialized AsyncRecords object
        """
        if self._materialized is None:
            self._materialized = await self._read_func()
        return self._materialized

    async def __aiter__(self) -> AsyncIterator[dict[str, object]]:
        """Make AsyncLazyRecords async iterable (auto-materializes)."""
        async for row in await self.collect():
            yield row

    async def rows(self) -> List[dict[str, object]]:
        """Return materialized list of all rows (auto-materializes).

        Returns:
            List of row dictionaries
        """
        return await (await self.collect()).rows()

    async def iter(self) -> AsyncIterator[dict[str, object]]:
        """Return an async iterator over rows (auto-materializes).

        Returns:
            AsyncIterator of row dictionaries
        """
        async for row in await self.collect():
            yield row

    @property
    def schema(self) -> Optional[Sequence["ColumnDef"]]:
        """Get the schema for these records.

        Returns:
            Schema if available, None otherwise
        """
        # Try to get schema without materializing if possible
        if self._schema is not None:
            return self._schema
        # Otherwise would need to materialize, but property can't be async
        # So return None and let materialized Records provide schema
        return None

    async def select(self, *columns: str) -> "AsyncRecords":
        """Select specific columns from records (auto-materializes).

        Args:
            *columns: Column names to select. Must be strings.

        Returns:
            New AsyncRecords with selected columns

        Example:
            >>> async_lazy_records = AsyncLazyRecords(_read_func=lambda: AsyncRecords(_data=[{"id": 1, "name": "Alice"}]))
            >>> selected = await async_lazy_records.select("id")
            >>> async for row in selected:
            ...     print(row)
            {"id": 1}
        """
        return await (await self.collect()).select(*columns)

    async def rename(
        self, columns: Union[Dict[str, str], str], new_name: Optional[str] = None
    ) -> "AsyncRecords":
        """Rename columns in records (auto-materializes).

        Args:
            columns: Either a dict mapping old_name -> new_name, or a single column name
            new_name: New name for the column (required if columns is a string)

        Returns:
            New AsyncRecords with renamed columns

        Example:
            >>> async_lazy_records = AsyncLazyRecords(_read_func=lambda: AsyncRecords(_data=[{"id": 1}]))
            >>> renamed = await async_lazy_records.rename("id", "user_id")
            >>> async for row in renamed:
            ...     print(row)
            {"user_id": 1}
        """
        return await (await self.collect()).rename(columns, new_name)

    async def head(self, n: int = 5) -> List[dict[str, object]]:
        """Return first n rows as list (auto-materializes).

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of first n row dictionaries
        """
        return await (await self.collect()).head(n)

    async def tail(self, n: int = 5) -> List[dict[str, object]]:
        """Return last n rows as list (auto-materializes).

        Args:
            n: Number of rows to return (default: 5)

        Returns:
            List of last n row dictionaries
        """
        return await (await self.collect()).tail(n)

    async def first(self) -> Optional[dict[str, object]]:
        """Return first row or None if empty (auto-materializes).

        Returns:
            First row dictionary or None if AsyncLazyRecords is empty
        """
        return await (await self.collect()).first()

    async def last(self) -> Optional[dict[str, object]]:
        """Return last row or None if empty (auto-materializes).

        Returns:
            Last row dictionary or None if AsyncLazyRecords is empty
        """
        return await (await self.collect()).last()

    async def insert_into(self, table: Union[str, "AsyncTableHandle"]) -> int:
        """Insert records into a table (auto-materializes).

        Args:
            table: Table name (str) or AsyncTableHandle

        Returns:
            Number of rows inserted

        Raises:
            RuntimeError: If no database is attached
        """
        return await (await self.collect()).insert_into(table)



"""Async Parquet file reader implementation."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncIterator, Callable, Dict, List, Optional, Sequence, cast

# aiofiles is not directly used here, but required for async file operations
# The import check is handled by the caller

from ...io.records import AsyncRecords
from ...table.schema import ColumnDef

if TYPE_CHECKING:
    from ...table.async_table import AsyncDatabase


async def read_parquet(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
) -> AsyncRecords:
    """Read Parquet file asynchronously and return AsyncRecords.

    Args:
        path: Path to Parquet file
        database: AsyncDatabase instance
        schema: Optional explicit schema (currently unused, schema from Parquet file is used)
        options: Reader options (unused for Parquet)

    Returns:
        AsyncRecords containing the Parquet data

    Raises:
        FileNotFoundError: If file doesn't exist
        RuntimeError: If pandas or pyarrow are not installed
    """
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"Parquet file not found: {path}")

    # Extract Parquet options
    merge_schema = cast(bool, options.get("mergeSchema", False))

    # Note: rebaseDatetimeInRead, datetimeRebaseMode, and int96RebaseMode options
    # are accepted but not yet implemented (PyArrow handles rebasing automatically)

    # For async, we'll read the file content first, then parse
    # Note: pyarrow doesn't have native async support, so we use asyncio.to_thread
    import asyncio

    def _read_parquet_sync() -> List[Dict[str, object]]:
        from ...utils.optional_deps import get_pyarrow_parquet

        pq = get_pyarrow_parquet(required=True)

        read_options: Dict[str, object] = {}
        if merge_schema:
            # Note: PyArrow's read_table doesn't directly support mergeSchema
            # This would require reading multiple files and merging schemas
            # For single file, this is a no-op
            pass

        table = pq.read_table(str(path_obj), **read_options)
        from ...utils.optional_deps import get_pandas

        get_pandas(required=True)
        df = table.to_pandas()
        return df.to_dict("records")  # type: ignore[no-any-return]

    rows = await asyncio.to_thread(_read_parquet_sync)

    if not rows:
        if schema:
            return _create_async_records_from_schema(database, schema, [])
        return _create_async_records_from_data(database, [], schema)

    # Infer or use explicit schema
    if schema:
        final_schema = schema
    else:
        from .schema_inference import infer_schema_from_rows

        final_schema = infer_schema_from_rows(rows)

    # Convert values to appropriate types
    from .schema_inference import apply_schema_to_rows

    typed_rows = apply_schema_to_rows(rows, final_schema)

    return _create_async_records_from_data(database, typed_rows, final_schema)


async def read_parquet_stream(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
) -> AsyncRecords:
    """Read Parquet file asynchronously in streaming mode (row group by row group).

    Args:
        path: Path to Parquet file
        database: AsyncDatabase instance
        schema: Optional explicit schema
        options: Reader options (unused for Parquet)

    Returns:
        AsyncRecords with streaming generator

    Raises:
        FileNotFoundError: If file doesn't exist
        RuntimeError: If pyarrow is not installed
    """
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"Parquet file not found: {path}")

    import asyncio
    from ...utils.optional_deps import get_pyarrow_parquet

    pq = get_pyarrow_parquet(required=True)

    def _get_parquet_file() -> Any:
        return pq.ParquetFile(str(path_obj))

    parquet_file = await asyncio.to_thread(_get_parquet_file)

    async def _chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        for i in range(parquet_file.num_row_groups):

            def _read_row_group(idx: int) -> List[Dict[str, object]]:
                from ...utils.optional_deps import get_pandas

                row_group = parquet_file.read_row_group(idx)
                get_pandas(required=True)
                df = row_group.to_pandas()
                return df.to_dict("records")  # type: ignore[no-any-return]

            rows = await asyncio.to_thread(_read_row_group, i)
            if rows:
                yield rows

    # Read first row group to infer schema
    first_chunk_gen = _chunk_generator()
    try:
        first_chunk = await first_chunk_gen.__anext__()
    except StopAsyncIteration:
        if schema:
            return _create_async_records_from_schema(database, schema, [])
        return _create_async_records_from_data(database, [], schema)

    # Infer or use explicit schema
    if schema:
        final_schema = schema
    else:
        from .schema_inference import infer_schema_from_rows

        final_schema = infer_schema_from_rows(first_chunk)

    from .schema_inference import apply_schema_to_rows

    async def _typed_chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        yield apply_schema_to_rows(first_chunk, final_schema)
        async for chunk in first_chunk_gen:
            yield apply_schema_to_rows(chunk, final_schema)

    return _create_async_records_from_stream(database, _typed_chunk_generator, final_schema)


def _create_async_records_from_data(
    database: "AsyncDatabase", rows: List[Dict[str, object]], schema: Optional[Sequence[ColumnDef]]
) -> AsyncRecords:
    """Create AsyncRecords from materialized data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_schema(
    database: "AsyncDatabase", schema: Sequence[ColumnDef], rows: List[Dict[str, object]]
) -> AsyncRecords:
    """Create AsyncRecords with explicit schema but no data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_stream(
    database: "AsyncDatabase",
    chunk_generator: Callable[[], AsyncIterator[List[Dict[str, object]]]],
    schema: Sequence[ColumnDef],
) -> AsyncRecords:
    """Create AsyncRecords from streaming generator."""
    return AsyncRecords(_generator=chunk_generator, _schema=schema, _database=database)



"""Schema definition primitives for table creation."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class ColumnDef:
    """Definition of a single table column."""

    name: str
    type_name: str
    nullable: bool = True
    default: object | None = None
    primary_key: bool = False
    precision: int | None = None  # For DECIMAL/NUMERIC types
    scale: int | None = None  # For DECIMAL/NUMERIC types


@dataclass(frozen=True)
class UniqueConstraint:
    """Definition of a UNIQUE constraint."""

    name: Optional[str] = None
    columns: Union[str, Sequence[str]] = ()

    def __post_init__(self) -> None:
        """Validate constraint definition."""
        if not self.columns:
            raise ValueError("UniqueConstraint must specify at least one column")


@dataclass(frozen=True)
class CheckConstraint:
    """Definition of a CHECK constraint."""

    name: Optional[str] = None
    expression: str = ""  # SQL expression string

    def __post_init__(self) -> None:
        """Validate constraint definition."""
        if not self.expression:
            raise ValueError("CheckConstraint must specify an expression")


@dataclass(frozen=True)
class ForeignKeyConstraint:
    """Definition of a FOREIGN KEY constraint."""

    name: Optional[str] = None
    columns: Union[str, Sequence[str]] = ()
    references_table: str = ""
    references_columns: Union[str, Sequence[str]] = ()
    on_delete: Optional[str] = None  # e.g., "CASCADE", "SET NULL", "RESTRICT"
    on_update: Optional[str] = None  # e.g., "CASCADE", "SET NULL", "RESTRICT"


@dataclass(frozen=True)
class TableSchema:
    """Complete schema definition for a table."""

    name: str
    columns: Sequence[ColumnDef]
    if_not_exists: bool = True
    temporary: bool = False
    constraints: Sequence[Union[UniqueConstraint, CheckConstraint, ForeignKeyConstraint]] = ()


def column(
    name: str,
    type_name: str,
    nullable: bool = True,
    default: object | None = None,
    primary_key: bool = False,
    precision: int | None = None,
    scale: int | None = None,
) -> ColumnDef:
    """Convenience helper for creating column definitions."""
    return ColumnDef(
        name=name,
        type_name=type_name,
        nullable=nullable,
        default=default,
        primary_key=primary_key,
        precision=precision,
        scale=scale,
    )


def decimal(
    name: str,
    precision: int,
    scale: int = 0,
    nullable: bool = True,
    default: object | None = None,
    primary_key: bool = False,
) -> ColumnDef:
    """Convenience helper for creating DECIMAL/NUMERIC column definitions.

    Args:
        name: Column name
        precision: Total number of digits
        scale: Number of digits after the decimal point
        nullable: Whether the column can be NULL
        default: Default value for the column
        primary_key: Whether this column is a primary key

    Returns:
        ColumnDef with type_name="DECIMAL" and precision/scale set

    Example:
        >>> from moltres.table.schema import decimal
        >>> col = decimal("price", precision=10, scale=2)  # DECIMAL(10, 2)
    """
    return ColumnDef(
        name=name,
        type_name="DECIMAL",
        nullable=nullable,
        default=default,
        primary_key=primary_key,
        precision=precision,
        scale=scale,
    )


def uuid(
    name: str,
    nullable: bool = True,
    default: object | None = None,
    primary_key: bool = False,
) -> ColumnDef:
    """Convenience helper for creating UUID column definitions.

    Args:
        name: Column name
        nullable: Whether the column can be NULL
        default: Default value for the column
        primary_key: Whether this column is a primary key

    Returns:
        ColumnDef with type_name="UUID" (PostgreSQL) or "CHAR(36)" (MySQL) or "TEXT" (SQLite)

    Example:
        >>> from moltres.table.schema import uuid
        >>> col = uuid("id", primary_key=True)  # UUID type
    """
    return ColumnDef(
        name=name,
        type_name="UUID",
        nullable=nullable,
        default=default,
        primary_key=primary_key,
    )


def json(
    name: str,
    nullable: bool = True,
    default: object | None = None,
    jsonb: bool = False,
) -> ColumnDef:
    """Convenience helper for creating JSON/JSONB column definitions.

    Args:
        name: Column name
        nullable: Whether the column can be NULL
        default: Default value for the column
        jsonb: If True, use JSONB (PostgreSQL only), otherwise use JSON

    Returns:
        ColumnDef with type_name="JSONB" (PostgreSQL with jsonb=True), "JSON" (MySQL/PostgreSQL), or "TEXT" (SQLite)

    Example:
        >>> from moltres.table.schema import json
        >>> col = json("data")  # JSON type
        >>> col2 = json("metadata", jsonb=True)  # JSONB type (PostgreSQL)
    """
    type_name = "JSONB" if jsonb else "JSON"
    return ColumnDef(
        name=name,
        type_name=type_name,
        nullable=nullable,
        default=default,
        primary_key=False,  # JSON columns typically aren't primary keys
    )


def unique(columns: Union[str, Sequence[str]], name: Optional[str] = None) -> UniqueConstraint:
    """Convenience helper for creating UNIQUE constraints.

    Args:
        columns: Column name(s) for the unique constraint
        name: Optional constraint name

    Returns:
        UniqueConstraint object

    Example:
        >>> from moltres.table.schema import unique
        >>> # Single column unique constraint
        >>> uq1 = unique("email")
        >>> # Multi-column unique constraint
        >>> uq2 = unique(["user_id", "session_id"], name="uq_user_session")
    """
    if not columns:
        raise ValueError("UniqueConstraint must specify at least one column")
    # Normalize to tuple for consistency
    if isinstance(columns, str):
        columns = (columns,)
    elif isinstance(columns, Sequence):
        columns = tuple(columns)
    return UniqueConstraint(name=name, columns=columns)


def check(expression: str, name: Optional[str] = None) -> CheckConstraint:
    """Convenience helper for creating CHECK constraints.

    Args:
        expression: SQL expression for the check constraint (e.g., "age > 0")
        name: Optional constraint name

    Returns:
        CheckConstraint object

    Example:
        >>> from moltres.table.schema import check
        >>> ck = check("age >= 0 AND age <= 150", name="ck_valid_age")
    """
    return CheckConstraint(name=name, expression=expression)


def foreign_key(
    columns: Union[str, Sequence[str]],
    references_table: str,
    references_columns: Union[str, Sequence[str]],
    name: Optional[str] = None,
    on_delete: Optional[str] = None,
    on_update: Optional[str] = None,
) -> ForeignKeyConstraint:
    """Convenience helper for creating FOREIGN KEY constraints.

    Args:
        columns: Column name(s) in this table
        references_table: Name of the referenced table
        references_columns: Column name(s) in the referenced table
        name: Optional constraint name
        on_delete: Optional action on delete (e.g., "CASCADE", "SET NULL", "RESTRICT")
        on_update: Optional action on update (e.g., "CASCADE", "SET NULL", "RESTRICT")

    Returns:
        ForeignKeyConstraint object

    Example:
        >>> from moltres.table.schema import foreign_key
        >>> # Single column foreign key
        >>> fk1 = foreign_key("user_id", "users", "id", on_delete="CASCADE")
        >>> # Multi-column foreign key
        >>> fk2 = foreign_key(["order_id", "item_id"], "order_items", ["id", "id"])
    """
    if not columns:
        raise ValueError("ForeignKeyConstraint must specify at least one column")
    if not references_table:
        raise ValueError("ForeignKeyConstraint must specify a references_table")
    if not references_columns:
        raise ValueError("ForeignKeyConstraint must specify at least one references_column")
    # Normalize to tuples for consistency
    if isinstance(columns, str):
        columns = (columns,)
    elif isinstance(columns, Sequence):
        columns = tuple(columns)
    if isinstance(references_columns, str):
        references_columns = (references_columns,)
    elif isinstance(references_columns, Sequence):
        references_columns = tuple(references_columns)
    # Validate column counts match
    if len(columns) != len(references_columns):
        raise ValueError(
            f"ForeignKeyConstraint columns count ({len(columns)}) "
            f"must match references_columns count ({len(references_columns)})"
        )
    return ForeignKeyConstraint(
        name=name,
        columns=columns,
        references_table=references_table,
        references_columns=references_columns,
        on_delete=on_delete,
        on_update=on_update,
    )



"""Async data loading operations."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence

from ..io.records import AsyncLazyRecords, AsyncRecords
from ..logical.operators import file_scan
from ..table.schema import ColumnDef
from .async_dataframe import AsyncDataFrame
from .readers.async_csv_reader import read_csv, read_csv_stream
from .readers.async_json_reader import (
    read_json,
    read_json_stream,
    read_jsonl,
    read_jsonl_stream,
)
from .readers.async_text_reader import read_text, read_text_stream


# Lazy import for parquet (optional dependency)
def _get_parquet_readers() -> tuple[Any, Any]:
    """Lazy import for parquet readers."""
    try:
        from .readers.async_parquet_reader import read_parquet, read_parquet_stream

        return read_parquet, read_parquet_stream
    except ImportError:
        return None, None


if TYPE_CHECKING:
    from ..table.async_table import AsyncDatabase


class AsyncDataLoader:
    """Builder for loading data from files and tables as AsyncDataFrames."""

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._schema: Optional[Sequence[ColumnDef]] = None
        self._options: Dict[str, object] = {}

    def stream(self, enabled: bool = True) -> "AsyncDataLoader":
        """Enable or disable streaming mode (chunked reading for large files)."""
        self._options["stream"] = enabled
        return self

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncDataLoader":
        """Set an explicit schema for the data source."""
        self._schema = schema
        return self

    def option(self, key: str, value: object) -> "AsyncDataLoader":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        self._options[key] = value
        return self

    def options(self, **options: object) -> "AsyncDataLoader":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining

        Example:
            >>> df = await db.read.options(header=True, delimiter=",").csv("data.csv")
        """
        self._options.update(options)
        return self

    async def table(self, name: str) -> AsyncDataFrame:
        """Read from a database table as an AsyncDataFrame.

        Note: This is equivalent to await db.table(name).select().
        Returns an AsyncDataFrame that can be transformed before execution.
        """
        table_handle = await self._database.table(name)
        return table_handle.select()

    async def csv(self, path: str) -> AsyncDataFrame:
        """Read a CSV file as an AsyncDataFrame.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncDataFrame containing the CSV data (lazy, materialized on .collect())
        """
        plan = file_scan(
            path=path,
            format="csv",
            schema=self._schema,
            options=self._options,
        )
        return AsyncDataFrame(plan=plan, database=self._database)

    async def json(self, path: str) -> AsyncDataFrame:
        """Read a JSON file (array of objects) as an AsyncDataFrame.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncDataFrame containing the JSON data (lazy, materialized on .collect())
        """
        plan = file_scan(
            path=path,
            format="json",
            schema=self._schema,
            options=self._options,
        )
        return AsyncDataFrame(plan=plan, database=self._database)

    async def jsonl(self, path: str) -> AsyncDataFrame:
        """Read a JSONL file (one JSON object per line) as an AsyncDataFrame.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncDataFrame containing the JSONL data (lazy, materialized on .collect())
        """
        plan = file_scan(
            path=path,
            format="jsonl",
            schema=self._schema,
            options=self._options,
        )
        return AsyncDataFrame(plan=plan, database=self._database)

    async def parquet(self, path: str) -> AsyncDataFrame:
        """Read a Parquet file as an AsyncDataFrame.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncDataFrame containing the Parquet data (lazy, materialized on .collect())

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        plan = file_scan(
            path=path,
            format="parquet",
            schema=self._schema,
            options=self._options,
        )
        return AsyncDataFrame(plan=plan, database=self._database)

    async def text(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (one line per row) as an AsyncDataFrame.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        plan = file_scan(
            path=path,
            format="text",
            schema=self._schema,
            options=self._options,
            column_name=column_name,
        )
        return AsyncDataFrame(plan=plan, database=self._database)

    async def textFile(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (PySpark-compatible alias for text()).

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self.text(path, column_name)

    async def format(self, source: str) -> "AsyncFormatReader":
        """Specify the data source format.

        Args:
            source: Format name (e.g., "csv", "json", "parquet")

        Returns:
            AsyncFormatReader for the specified format
        """
        return AsyncFormatReader(self, source)


class AsyncFormatReader:
    """Builder for format-specific async reads."""

    def __init__(self, reader: AsyncDataLoader, source: str):
        self._reader = reader
        self._source = source.lower()

    async def load(self, path: str) -> AsyncDataFrame:
        """Load data from the specified path using the configured format.

        Args:
            path: Path to the data file

        Returns:
            AsyncDataFrame containing the data (lazy, materialized on .collect())

        Raises:
            ValueError: If format is unsupported
        """
        if self._source == "csv":
            return await self._reader.csv(path)
        elif self._source == "json":
            return await self._reader.json(path)
        elif self._source == "jsonl":
            return await self._reader.jsonl(path)
        elif self._source == "parquet":
            read_parquet, _ = _get_parquet_readers()
            if read_parquet is None:
                raise ImportError(
                    "Parquet support requires pyarrow. Install with: pip install pyarrow"
                )
            return await self._reader.parquet(path)
        elif self._source == "text":
            return await self._reader.text(path)
        else:
            raise ValueError(f"Unsupported format: {self._source}")


class AsyncRecordsLoader:
    """Builder for loading data from files as AsyncLazyRecords (lazy AsyncRecords).

    Provides backward compatibility and convenience for cases where AsyncRecords are preferred
    over AsyncDataFrames. Use await db.read.records.csv() etc. to get AsyncLazyRecords directly.
    AsyncLazyRecords materialize on-demand when used.
    """

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._schema: Optional[Sequence[ColumnDef]] = None
        self._options: Dict[str, object] = {}

    def stream(self, enabled: bool = True) -> "AsyncRecordsLoader":
        """Enable or disable streaming mode (chunked reading for large files)."""
        self._options["stream"] = enabled
        return self

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncRecordsLoader":
        """Set an explicit schema for the data source."""
        self._schema = schema
        return self

    def option(self, key: str, value: object) -> "AsyncRecordsLoader":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        self._options[key] = value
        return self

    def options(self, **options: object) -> "AsyncRecordsLoader":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining
        """
        self._options.update(options)
        return self

    def csv(self, path: str) -> AsyncLazyRecords:
        """Read a CSV file as AsyncLazyRecords.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncLazyRecords containing the CSV data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_csv_stream(path, self._database, self._schema, self._options)
            return await read_csv(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def json(self, path: str) -> AsyncLazyRecords:
        """Read a JSON file (array of objects) as AsyncLazyRecords.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncLazyRecords containing the JSON data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_json_stream(path, self._database, self._schema, self._options)
            return await read_json(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def jsonl(self, path: str) -> AsyncLazyRecords:
        """Read a JSONL file (one JSON object per line) as AsyncLazyRecords.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncLazyRecords containing the JSONL data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_jsonl_stream(path, self._database, self._schema, self._options)
            return await read_jsonl(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def parquet(self, path: str) -> AsyncLazyRecords:
        """Read a Parquet file as AsyncLazyRecords.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncLazyRecords containing the Parquet data (materializes on-demand)

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        read_parquet, read_parquet_stream = _get_parquet_readers()
        if read_parquet is None:
            raise ImportError("Parquet support requires pyarrow. Install with: pip install pyarrow")

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_parquet_stream(path, self._database, self._schema, self._options)  # type: ignore[no-any-return]
            return await read_parquet(path, self._database, self._schema, self._options)  # type: ignore[no-any-return]

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def text(self, path: str, column_name: str = "value") -> AsyncLazyRecords:
        """Read a text file as a single column (one line per row) as AsyncLazyRecords.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncLazyRecords containing the text file lines (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_text_stream(
                    path, self._database, self._schema, self._options, column_name
                )
            return await read_text(path, self._database, self._schema, self._options, column_name)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def dicts(self, data: Sequence[Dict[str, object]]) -> AsyncRecords:
        """Create AsyncRecords from a list of dictionaries.

        Note: This returns AsyncRecords (not AsyncLazyRecords) since the data is already materialized.

        Args:
            data: List of dictionaries to convert to AsyncRecords

        Returns:
            AsyncRecords containing the data (already materialized)
        """
        return AsyncRecords(_data=list(data), _database=self._database, _schema=self._schema)


class AsyncReadAccessor:
    """Accessor for async read operations.

    Provides PySpark-style API: await db.read.table(), await db.read.csv(), etc.
    Also provides backward compatibility via db.read.records.*
    """

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._loader = AsyncDataLoader(database)
        self._records = AsyncRecordsLoader(database)

    @property
    def records(self) -> AsyncRecordsLoader:
        """Access to AsyncRecords-based read methods."""
        return self._records

    # Builder methods that configure the underlying AsyncDataLoader
    def stream(self, enabled: bool = True) -> "AsyncReadAccessor":
        """Enable or disable streaming mode (chunked reading for large files)."""
        self._loader.stream(enabled)
        return self

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncReadAccessor":
        """Set an explicit schema for the data source."""
        self._loader.schema(schema)
        return self

    def option(self, key: str, value: object) -> "AsyncReadAccessor":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        self._loader.option(key, value)
        return self

    def options(self, **options: object) -> "AsyncReadAccessor":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining

        Example:
            >>> df = await db.read.options(header=True, delimiter=",").csv("data.csv")
        """
        self._loader.options(**options)
        return self

    # AsyncDataFrame read methods (delegate to AsyncDataLoader)
    async def table(self, name: str) -> AsyncDataFrame:
        """Read from a database table as an AsyncDataFrame.

        Args:
            name: Name of the table to read

        Returns:
            AsyncDataFrame that can be transformed before execution

        Example:
            >>> df = await db.read.table("users")
            >>> results = await df.collect()
        """
        return await self._loader.table(name)

    async def csv(self, path: str) -> AsyncDataFrame:
        """Read a CSV file as an AsyncDataFrame.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncDataFrame containing the CSV data (lazy, materialized on .collect())
        """
        return await self._loader.csv(path)

    async def json(self, path: str) -> AsyncDataFrame:
        """Read a JSON file (array of objects) as an AsyncDataFrame.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncDataFrame containing the JSON data (lazy, materialized on .collect())
        """
        return await self._loader.json(path)

    async def jsonl(self, path: str) -> AsyncDataFrame:
        """Read a JSONL file (one JSON object per line) as an AsyncDataFrame.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncDataFrame containing the JSONL data (lazy, materialized on .collect())
        """
        return await self._loader.jsonl(path)

    async def parquet(self, path: str) -> AsyncDataFrame:
        """Read a Parquet file as an AsyncDataFrame.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncDataFrame containing the Parquet data (lazy, materialized on .collect())

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        return await self._loader.parquet(path)

    async def text(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (one line per row) as an AsyncDataFrame.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self._loader.text(path, column_name)

    async def textFile(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (PySpark-compatible alias for text()).

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self._loader.textFile(path, column_name)

    async def format(self, source: str) -> AsyncFormatReader:
        """Specify the data source format.

        Args:
            source: Format name (e.g., "csv", "json", "parquet")

        Returns:
            AsyncFormatReader for the specified format
        """
        return await self._loader.format(source)



"""Structured logging and metrics hooks for Moltres operations."""

from __future__ import annotations

import json
import logging
import time
from typing import Any, Callable, Dict, Optional

logger = logging.getLogger(__name__)


class StructuredLogger:
    """Structured logger for Moltres operations.

    Provides structured JSON logging for queries, errors, and performance metrics.
    """

    def __init__(self, logger_instance: Optional[logging.Logger] = None):
        """Initialize structured logger.

        Args:
            logger_instance: Optional logger instance. Defaults to module logger.
        """
        self.logger = logger_instance or logger

    def log_query_start(
        self,
        sql: str,
        params: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Log query start event.

        Args:
            sql: SQL query string
            params: Query parameters
            metadata: Additional metadata
        """
        event = {
            "event": "query_start",
            "sql": sql[:500],  # Truncate long queries
            "timestamp": time.time(),
        }
        if params:
            event["params"] = params
        if metadata:
            event.update(metadata)
        self.logger.debug(json.dumps(event))

    def log_query_end(
        self,
        sql: str,
        duration: float,
        rowcount: Optional[int] = None,
        params: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Log query end event.

        Args:
            sql: SQL query string
            duration: Query duration in seconds
            rowcount: Number of rows returned
            params: Query parameters
            metadata: Additional metadata
        """
        event = {
            "event": "query_end",
            "sql": sql[:500],  # Truncate long queries
            "duration": duration,
            "timestamp": time.time(),
        }
        if rowcount is not None:
            event["rowcount"] = rowcount
        if params:
            event["params"] = params
        if metadata:
            event.update(metadata)

        # Log at appropriate level based on duration
        if duration > 5.0:
            self.logger.warning(json.dumps(event))
        elif duration > 1.0:
            self.logger.info(json.dumps(event))
        else:
            self.logger.debug(json.dumps(event))

    def log_query_error(
        self,
        sql: str,
        error: Exception,
        duration: Optional[float] = None,
        params: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Log query error event.

        Args:
            sql: SQL query string
            error: Exception that occurred
            duration: Query duration before error (if available)
            params: Query parameters
            metadata: Additional metadata
        """
        event = {
            "event": "query_error",
            "sql": sql[:500],  # Truncate long queries
            "error_type": type(error).__name__,
            "error_message": str(error),
            "timestamp": time.time(),
        }
        if duration is not None:
            event["duration"] = duration
        if params:
            event["params"] = params
        if metadata:
            event.update(metadata)
        self.logger.error(json.dumps(event), exc_info=error)

    def log_connection_event(
        self,
        event_type: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Log connection-related event.

        Args:
            event_type: Type of connection event (e.g., "connect", "disconnect", "pool_checkout")
            metadata: Additional metadata
        """
        event = {
            "event": "connection",
            "connection_event": event_type,
            "timestamp": time.time(),
        }
        if metadata:
            event.update(metadata)
        self.logger.debug(json.dumps(event))


class MetricsCollector:
    """Metrics collector for Moltres operations.

    Collects metrics for queries, connections, and errors.
    """

    def __init__(self) -> None:
        """Initialize metrics collector."""
        self._query_count = 0
        self._query_duration_sum = 0.0
        self._query_duration_max = 0.0
        self._error_count = 0
        self._connection_count = 0

    def record_query(self, duration: float, success: bool = True) -> None:
        """Record a query execution.

        Args:
            duration: Query duration in seconds
            success: Whether query succeeded
        """
        self._query_count += 1
        self._query_duration_sum += duration
        self._query_duration_max = max(self._query_duration_max, duration)
        if not success:
            self._error_count += 1

    def record_connection(self) -> None:
        """Record a connection event."""
        self._connection_count += 1

    def get_metrics(self) -> Dict[str, Any]:
        """Get current metrics.

        Returns:
            Dictionary of metrics
        """
        avg_duration = (
            self._query_duration_sum / self._query_count if self._query_count > 0 else 0.0
        )
        return {
            "query_count": self._query_count,
            "query_duration_avg": avg_duration,
            "query_duration_max": self._query_duration_max,
            "error_count": self._error_count,
            "error_rate": (self._error_count / self._query_count if self._query_count > 0 else 0.0),
            "connection_count": self._connection_count,
        }

    def reset(self) -> None:
        """Reset all metrics."""
        self._query_count = 0
        self._query_duration_sum = 0.0
        self._query_duration_max = 0.0
        self._error_count = 0
        self._connection_count = 0


# Global instances
_structured_logger = StructuredLogger()
_metrics_collector = MetricsCollector()


def get_structured_logger() -> StructuredLogger:
    """Get the global structured logger instance."""
    return _structured_logger


def get_metrics_collector() -> MetricsCollector:
    """Get the global metrics collector instance."""
    return _metrics_collector


def create_performance_hook_from_logger() -> Callable[[str, float, Dict[str, Any]], None]:
    """Create a performance hook that uses structured logging.

    Returns:
        Callback function suitable for register_performance_hook
    """

    def hook(sql: str, elapsed: float, metadata: Dict[str, Any]) -> None:
        event = metadata.get("event", "query_end")
        if event == "query_start":
            _structured_logger.log_query_start(
                sql,
                params=metadata.get("params"),
                metadata={k: v for k, v in metadata.items() if k not in ("event", "params")},
            )
        elif event == "query_end":
            _structured_logger.log_query_end(
                sql,
                elapsed,
                rowcount=metadata.get("rowcount"),
                params=metadata.get("params"),
                metadata={
                    k: v for k, v in metadata.items() if k not in ("event", "rowcount", "params")
                },
            )
            _metrics_collector.record_query(elapsed, success=True)
        elif event == "query_error":
            _structured_logger.log_query_error(
                sql,
                metadata.get("error", Exception("Unknown error")),
                duration=elapsed,
                params=metadata.get("params"),
                metadata={
                    k: v
                    for k, v in metadata.items()
                    if k not in ("event", "error", "duration", "params")
                },
            )
            _metrics_collector.record_query(elapsed, success=False)

    return hook



"""Async CSV file reader implementation."""

from __future__ import annotations

import csv
import io
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    cast,
)

from ...io.records import AsyncRecords
from ...table.schema import ColumnDef
from .compression import read_compressed_async

if TYPE_CHECKING:
    from ...table.async_table import AsyncDatabase


async def read_csv(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
) -> AsyncRecords:
    """Read CSV file asynchronously and return AsyncRecords.

    Args:
        path: Path to CSV file
        database: AsyncDatabase instance
        schema: Optional explicit schema
        options: Reader options (header, delimiter, inferSchema)

    Returns:
        AsyncRecords containing the CSV data

    Raises:
        FileNotFoundError: If file doesn't exist
        ValueError: If file is empty or invalid
    """
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"CSV file not found: {path}")

    # Extract all options with defaults (same as sync read_csv)
    header = cast(bool, options.get("header", True))
    delimiter = cast(str, options.get("delimiter", options.get("sep", ",")))
    infer_schema = cast(bool, options.get("inferSchema", True))
    compression = cast(Optional[str], options.get("compression", None))
    encoding = cast(str, options.get("encoding", "UTF-8"))
    quote = cast(str, options.get("quote", '"'))
    escape = cast(str, options.get("escape", "\\"))
    null_value = cast(str, options.get("nullValue", ""))
    nan_value = cast(str, options.get("nanValue", "NaN"))
    date_format = cast(Optional[str], options.get("dateFormat", None))
    timestamp_format = cast(Optional[str], options.get("timestampFormat", None))
    ignore_leading_whitespace = cast(bool, options.get("ignoreLeadingWhiteSpace", False))
    ignore_trailing_whitespace = cast(bool, options.get("ignoreTrailingWhiteSpace", False))
    comment = cast(Optional[str], options.get("comment", None))
    sampling_ratio = cast(float, options.get("samplingRatio", 1.0))
    mode = cast(str, options.get("mode", "PERMISSIVE"))
    corrupt_column = cast(Optional[str], options.get("columnNameOfCorruptRecord", None))

    rows: List[Dict[str, object]] = []
    corrupt_records: List[Dict[str, object]] = []

    content = await read_compressed_async(path, compression=compression, encoding=encoding)

    # Configure CSV reader with options
    csv_kwargs: Dict[str, Any] = {
        "delimiter": delimiter,
        "skipinitialspace": ignore_leading_whitespace or ignore_trailing_whitespace,
    }

    # Handle quote character
    if quote and len(quote) == 1:
        csv_kwargs["quotechar"] = quote
    else:
        csv_kwargs["quotechar"] = '"'

    # Handle escape character
    if escape and len(escape) == 1:
        csv_kwargs["escapechar"] = escape
    else:
        csv_kwargs["escapechar"] = None

    # Handle comment character
    if comment and len(comment) == 1:
        csv_kwargs["comment"] = comment

    def _process_value(value: str, col_name: Optional[str] = None) -> object:
        """Process a CSV value according to options."""
        if value is None:
            return None

        # Handle null values
        if value == null_value:
            return None

        # Handle NaN values
        if value == nan_value:
            return float("nan")

        # Handle whitespace
        if ignore_leading_whitespace:
            value = value.lstrip()
        if ignore_trailing_whitespace:
            value = value.rstrip()

        return value

    def _parse_row(row_data: List[str], column_names: List[str]) -> Optional[Dict[str, object]]:
        """Parse a CSV row with error handling based on mode."""
        try:
            row_dict: Dict[str, object] = {}
            for i, col_name in enumerate(column_names):
                raw_value = row_data[i] if i < len(row_data) else None
                processed_value: object = (
                    _process_value(raw_value, col_name) if raw_value is not None else None
                )
                row_dict[col_name] = processed_value

            # Add corrupt record column if specified
            if corrupt_column:
                row_dict[corrupt_column] = None

            return row_dict
        except Exception as e:
            if mode == "FAILFAST":
                raise ValueError(f"Failed to parse CSV row: {e}") from e
            elif mode == "DROPMALFORMED":
                return None
            else:  # PERMISSIVE
                if corrupt_column:
                    corrupt_row: Dict[str, object] = {corrupt_column: str(row_data)}
                    corrupt_records.append(corrupt_row)
                return None

    # Parse CSV content
    csv_file = io.StringIO(content)
    if header and not schema:
        # Use DictReader when we have headers and no explicit schema
        dict_reader: Any = csv.DictReader(csv_file, **csv_kwargs)
        column_names = dict_reader.fieldnames or []

        for row in dict_reader:
            try:
                processed_row: Dict[str, object] = {}
                for k, v in row.items():
                    processed_row[k] = _process_value(v, k)

                # Add corrupt record column if specified
                if corrupt_column and corrupt_column not in processed_row:
                    processed_row[corrupt_column] = None

                rows.append(processed_row)
            except Exception as e:
                if mode == "FAILFAST":
                    raise ValueError(f"Failed to parse CSV row: {e}") from e
                elif mode == "DROPMALFORMED":
                    continue
                else:  # PERMISSIVE
                    if corrupt_column:
                        corrupt_row: Dict[str, object] = {corrupt_column: str(row)}
                        corrupt_records.append(corrupt_row)
                    continue
    else:
        # Read without header or with explicit schema
        csv_reader: Any = csv.reader(csv_file, **csv_kwargs)
        if header and schema:
            # Skip header row when we have explicit schema
            next(csv_reader, None)

        if schema:
            # Use schema column names
            column_names = [col_def.name for col_def in schema]
            for row_data in csv_reader:
                if not row_data:  # Skip empty rows
                    continue
                parsed_row = _parse_row(row_data, column_names)
                if parsed_row is not None:
                    rows.append(parsed_row)
        else:
            # No header and no schema - can't determine column names
            raise ValueError(
                "CSV file without header requires explicit schema. Use .schema([ColumnDef(...), ...])"
            )

    if not rows:
        if schema:
            # Empty file with explicit schema
            return _create_async_records_from_schema(database, schema, [])
        raise ValueError(f"CSV file is empty: {path}")

    # Apply sampling ratio for schema inference
    sample_rows = rows
    if sampling_ratio < 1.0 and rows:
        sample_size = max(1, int(len(rows) * sampling_ratio))
        sample_rows = rows[:sample_size]

    # Infer or use explicit schema
    if schema:
        final_schema = schema
    elif infer_schema and sample_rows:
        from .schema_inference import infer_schema_from_rows

        final_schema = infer_schema_from_rows(
            sample_rows, date_format=date_format, timestamp_format=timestamp_format
        )
    elif rows:
        # All columns as TEXT
        final_schema = [
            ColumnDef(name=col, type_name="TEXT", nullable=True) for col in rows[0].keys()
        ]
    else:
        # Empty file
        final_schema = []

    # Add corrupt records column to schema if needed
    if corrupt_column and corrupt_column not in [col.name for col in final_schema]:
        final_schema = list(final_schema) + [
            ColumnDef(name=corrupt_column, type_name="TEXT", nullable=True)
        ]
        # Add None values for corrupt column to all existing rows
        for row in rows:
            if corrupt_column not in row:
                row[corrupt_column] = None
        # Add corrupt records
        rows.extend(corrupt_records)

    # Convert values to appropriate types based on schema
    from .schema_inference import apply_schema_to_rows

    typed_rows = apply_schema_to_rows(
        rows, final_schema, date_format=date_format, timestamp_format=timestamp_format
    )

    return _create_async_records_from_data(database, typed_rows, final_schema)


async def read_csv_stream(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
) -> AsyncRecords:
    """Read CSV file asynchronously in streaming mode (chunked).

    Args:
        path: Path to CSV file
        database: AsyncDatabase instance
        schema: Optional explicit schema
        options: Reader options (header, delimiter, inferSchema, chunk_size)

    Returns:
        AsyncRecords with streaming generator

    Raises:
        FileNotFoundError: If file doesn't exist
        ValueError: If file is empty or invalid
    """
    chunk_size = int(cast(int, options.get("chunk_size", 10000)))
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"CSV file not found: {path}")

    # Extract all options with defaults (same as sync read_csv_stream)
    header = cast(bool, options.get("header", True))
    delimiter = cast(str, options.get("delimiter", options.get("sep", ",")))
    infer_schema = cast(bool, options.get("inferSchema", True))
    compression = cast(Optional[str], options.get("compression", None))
    encoding = cast(str, options.get("encoding", "UTF-8"))
    quote = cast(str, options.get("quote", '"'))
    escape = cast(str, options.get("escape", "\\"))
    null_value = cast(str, options.get("nullValue", ""))
    nan_value = cast(str, options.get("nanValue", "NaN"))
    date_format = cast(Optional[str], options.get("dateFormat", None))
    timestamp_format = cast(Optional[str], options.get("timestampFormat", None))
    ignore_leading_whitespace = cast(bool, options.get("ignoreLeadingWhiteSpace", False))
    ignore_trailing_whitespace = cast(bool, options.get("ignoreTrailingWhiteSpace", False))
    comment = cast(Optional[str], options.get("comment", None))
    mode = cast(str, options.get("mode", "PERMISSIVE"))
    corrupt_column = cast(Optional[str], options.get("columnNameOfCorruptRecord", None))

    def _process_value(value: str, col_name: Optional[str] = None) -> object:
        """Process a CSV value according to options."""
        if value is None:
            return None

        # Handle null values
        if value == null_value:
            return None

        # Handle NaN values
        if value == nan_value:
            return float("nan")

        # Handle whitespace
        if ignore_leading_whitespace:
            value = value.lstrip()
        if ignore_trailing_whitespace:
            value = value.rstrip()

        return value

    async def _chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        # Configure CSV reader with options
        csv_kwargs: Dict[str, Any] = {
            "delimiter": delimiter,
            "skipinitialspace": ignore_leading_whitespace or ignore_trailing_whitespace,
        }

        # Handle quote character
        if quote and len(quote) == 1:
            csv_kwargs["quotechar"] = quote
        else:
            csv_kwargs["quotechar"] = '"'

        # Handle escape character
        if escape and len(escape) == 1:
            csv_kwargs["escapechar"] = escape
        else:
            csv_kwargs["escapechar"] = None

        # Handle comment character
        if comment and len(comment) == 1:
            csv_kwargs["comment"] = comment

        # Read and decompress file content
        content = await read_compressed_async(path, compression=compression, encoding=encoding)
        csv_file = io.StringIO(content)

        if header and not schema:
            reader = csv.DictReader(csv_file, **csv_kwargs)
            chunk = []
            for row in reader:
                try:
                    processed_row: Dict[str, object] = {}
                    for k, v in row.items():
                        processed_row[k] = _process_value(v, k)

                    # Add corrupt record column if specified
                    if corrupt_column and corrupt_column not in processed_row:
                        processed_row[corrupt_column] = None

                    chunk.append(processed_row)
                    if len(chunk) >= chunk_size:
                        yield chunk
                        chunk = []
                except Exception as e:
                    if mode == "FAILFAST":
                        raise ValueError(f"Failed to parse CSV row: {e}") from e
                    elif mode == "DROPMALFORMED":
                        continue
                    else:  # PERMISSIVE
                        if corrupt_column:
                            corrupt_row: Dict[str, object] = {corrupt_column: str(row)}
                            chunk.append(corrupt_row)
                        continue
            if chunk:
                yield chunk
        else:
            reader_obj: Any = csv.reader(csv_file, **csv_kwargs)
            if header and schema:
                next(reader_obj, None)
            if schema:
                column_names = [col_def.name for col_def in schema]
                chunk = []
                for row_data in reader_obj:
                    if not row_data:
                        continue
                    try:
                        row_dict: Dict[str, object] = {}
                        for i, col_name in enumerate(column_names):
                            raw_value = row_data[i] if i < len(row_data) else None
                            processed_value: object = (
                                _process_value(raw_value, col_name)
                                if raw_value is not None
                                else None
                            )
                            row_dict[col_name] = processed_value

                        # Add corrupt record column if specified
                        if corrupt_column:
                            row_dict[corrupt_column] = None

                        chunk.append(row_dict)
                        if len(chunk) >= chunk_size:
                            yield chunk
                            chunk = []
                    except Exception as e:
                        if mode == "FAILFAST":
                            raise ValueError(f"Failed to parse CSV row: {e}") from e
                        elif mode == "DROPMALFORMED":
                            continue
                        else:  # PERMISSIVE
                            if corrupt_column:
                                corrupt_record: Dict[str, object] = {corrupt_column: str(row_data)}
                                chunk.append(corrupt_record)
                            continue
                if chunk:
                    yield chunk
            else:
                raise ValueError(
                    "CSV file without header requires explicit schema. Use .schema([ColumnDef(...), ...])"
                )

    # Read first chunk to infer schema
    first_chunk_gen = _chunk_generator()
    try:
        first_chunk = await first_chunk_gen.__anext__()
    except StopAsyncIteration:
        if schema:
            return _create_async_records_from_schema(database, schema, [])
        raise ValueError(f"CSV file is empty: {path}")

    # Add corrupt records column to schema if needed
    if corrupt_column and first_chunk:
        if corrupt_column not in first_chunk[0]:
            # Add None values for corrupt column to all rows in first chunk
            for row in first_chunk:
                row[corrupt_column] = None

    # Infer or use explicit schema
    if schema:
        final_schema = schema
    elif infer_schema and first_chunk:
        from .schema_inference import infer_schema_from_rows

        final_schema = infer_schema_from_rows(
            first_chunk, date_format=date_format, timestamp_format=timestamp_format
        )
    elif first_chunk:
        final_schema = [
            ColumnDef(name=col, type_name="TEXT", nullable=True) for col in first_chunk[0].keys()
        ]
    else:
        final_schema = []

    # Ensure corrupt column is in schema
    if corrupt_column and corrupt_column not in [col.name for col in final_schema]:
        final_schema = list(final_schema) + [
            ColumnDef(name=corrupt_column, type_name="TEXT", nullable=True)
        ]

    # Create generator that applies schema and yields chunks
    from .schema_inference import apply_schema_to_rows

    async def _typed_chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        # Yield first chunk (already read)
        yield apply_schema_to_rows(
            first_chunk, final_schema, date_format=date_format, timestamp_format=timestamp_format
        )
        # Yield remaining chunks
        async for chunk in first_chunk_gen:
            yield apply_schema_to_rows(
                chunk, final_schema, date_format=date_format, timestamp_format=timestamp_format
            )

    return _create_async_records_from_stream(database, _typed_chunk_generator, final_schema)


def _create_async_records_from_data(
    database: "AsyncDatabase", rows: List[Dict[str, object]], schema: Optional[Sequence[ColumnDef]]
) -> AsyncRecords:
    """Create AsyncRecords from materialized data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_schema(
    database: "AsyncDatabase", schema: Sequence[ColumnDef], rows: List[Dict[str, object]]
) -> AsyncRecords:
    """Create AsyncRecords with explicit schema but no data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_stream(
    database: "AsyncDatabase",
    chunk_generator: Callable[[], AsyncIterator[List[Dict[str, object]]]],
    schema: Sequence[ColumnDef],
) -> AsyncRecords:
    """Create AsyncRecords from streaming generator."""
    return AsyncRecords(_generator=chunk_generator, _schema=schema, _database=database)



"""Grouped DataFrame helper."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Union

from ..expressions.column import Column, col
from ..logical import operators
from ..logical.plan import LogicalPlan
from .dataframe import DataFrame


@dataclass(frozen=True)
class GroupedDataFrame:
    """Represents a DataFrame grouped by one or more columns.

    This is returned by DataFrame.group_by() and provides aggregation methods.
    """

    plan: LogicalPlan
    keys: tuple[Column, ...]
    parent: DataFrame

    def agg(self, *aggregations: Union[Column, str, Dict[str, str]]) -> DataFrame:
        """Apply aggregation functions to the grouped data.

        Args:
            *aggregations: One or more aggregation expressions. Can be:
                - Column expressions (e.g., sum(col("amount")))
                - String column names (e.g., "amount" - defaults to sum())
                - Dictionary mapping column names to aggregation functions
                  (e.g., {"amount": "sum", "price": "avg"})

        Returns:
            DataFrame with aggregated results

        Raises:
            ValueError: If no aggregations are provided or if invalid
                aggregation expressions are used

        Example:
            >>> from moltres import col
            >>> from moltres.expressions import functions as F
            >>> # Using Column expressions
            >>> df.group_by("category").agg(
            ...     F.sum(col("amount")).alias("total"),
            ...     F.avg(col("price")).alias("avg_price"),
            ...     F.count("*").alias("count")
            ... )

            >>> # Using string column names (defaults to sum)
            >>> df.group_by("category").agg("amount", "price")

            >>> # Using dictionary syntax
            >>> df.group_by("category").agg({"amount": "sum", "price": "avg"})

            >>> # Mixed usage
            >>> df.group_by("category").agg("amount", F.sum(col("price")).alias("total_price"))
        """
        if not aggregations:
            raise ValueError("agg requires at least one aggregation expression")

        # Normalize all aggregations to Column expressions
        normalized_aggs = []
        for agg_expr in aggregations:
            if isinstance(agg_expr, str):
                # String column name - default to sum() and alias with column name
                from ..expressions.functions import sum as sum_func

                normalized_aggs.append(sum_func(col(agg_expr)).alias(agg_expr))
            elif isinstance(agg_expr, dict):
                # Dictionary syntax: {"column": "function"}
                for col_name, func_name in agg_expr.items():
                    agg_col = self._create_aggregation_from_string(col_name, func_name)
                    normalized_aggs.append(agg_col)
            elif isinstance(agg_expr, Column):
                # Already a Column expression
                normalized_aggs.append(agg_expr)
            else:
                raise ValueError(
                    f"Invalid aggregation type: {type(agg_expr)}. "
                    "Expected Column, str, or Dict[str, str]"
                )

        normalized = tuple(self._validate_aggregation(expr) for expr in normalized_aggs)
        plan = operators.aggregate(self.plan, self.keys, normalized)
        return DataFrame(plan=plan, database=self.parent.database)

    @staticmethod
    def _create_aggregation_from_string(column_name: str, func_name: str) -> Column:
        """Create an aggregation Column from a column name and function name string.

        Args:
            column_name: Name of the column to aggregate
            func_name: Name of the aggregation function (e.g., "sum", "avg", "min", "max", "count")

        Returns:
            Column expression for the aggregation

        Raises:
            ValueError: If the function name is not recognized
        """
        from ..expressions.functions import (
            avg,
            count,
            max as max_func,
            min as min_func,
            sum as sum_func,
            count_distinct,
        )

        from typing import Callable

        func_map: Dict[str, Callable[[Column], Column]] = {
            "sum": sum_func,
            "avg": avg,
            "average": avg,  # Alias for avg
            "min": min_func,
            "max": max_func,
            "count": count,
            "count_distinct": count_distinct,
        }

        func_name_lower = func_name.lower()
        if func_name_lower not in func_map:
            raise ValueError(
                f"Unknown aggregation function: {func_name}. "
                f"Supported functions: {', '.join(func_map.keys())}"
            )

        agg_func = func_map[func_name_lower]
        return agg_func(col(column_name)).alias(column_name)

    def pivot(
        self, pivot_col: str, values: Optional[Sequence[str]] = None
    ) -> "PivotedGroupedDataFrame":
        """Pivot the grouped data on a column.

        Args:
            pivot_col: Column to pivot on (values become column headers)
            values: Optional list of specific values to pivot (if None, must be provided later or discovered)

        Returns:
            PivotedGroupedDataFrame that can be aggregated

        Example:
            >>> df.group_by("category").pivot("status").agg("amount")
            >>> df.group_by("category").pivot("status", values=["active", "inactive"]).agg("amount")
        """
        return PivotedGroupedDataFrame(
            plan=self.plan,
            keys=self.keys,
            pivot_column=pivot_col,
            pivot_values=tuple(values) if values else None,
            parent=self.parent,
        )

    @staticmethod
    def _validate_aggregation(expr: Column) -> Column:
        """Validate that an expression is a valid aggregation.

        Args:
            expr: Column expression to validate

        Returns:
            The validated column expression

        Raises:
            ValueError: If the expression is not a valid aggregation
        """
        if not expr.op.startswith("agg_"):
            raise ValueError(
                "Aggregation expressions must be created with moltres aggregate helpers "
                "(e.g., sum(), avg(), count(), min(), max())"
            )
        return expr


@dataclass(frozen=True)
class PivotedGroupedDataFrame:
    """Represents a DataFrame grouped by columns with a pivot operation applied.

    This is returned by GroupedDataFrame.pivot() and provides aggregation methods
    that will create pivoted columns.
    """

    plan: LogicalPlan
    keys: tuple[Column, ...]
    pivot_column: str
    pivot_values: Optional[tuple[str, ...]]
    parent: DataFrame

    def agg(self, *aggregations: Union[Column, str, Dict[str, str]]) -> DataFrame:
        """Apply aggregation functions to the pivoted grouped data.

        Args:
            *aggregations: One or more aggregation expressions. Can be:
                - Column expressions (e.g., sum(col("amount")))
                - String column names (e.g., "amount" - defaults to sum())
                - Dictionary mapping column names to aggregation functions
                  (e.g., {"amount": "sum", "price": "avg"})

        Returns:
            DataFrame with pivoted aggregated results

        Raises:
            ValueError: If no aggregations are provided or if invalid
                aggregation expressions are used

        Example:
            >>> from moltres import col
            >>> from moltres.expressions import functions as F
            >>> # Using string column name
            >>> df.group_by("category").pivot("status").agg("amount")

            >>> # Using Column expression
            >>> df.group_by("category").pivot("status").agg(F.sum(col("amount")))

            >>> # With specific pivot values
            >>> df.group_by("category").pivot("status", values=["active", "inactive"]).agg("amount")
        """
        if not aggregations:
            raise ValueError("agg requires at least one aggregation expression")

        # Normalize all aggregations to Column expressions
        normalized_aggs = []
        for agg_expr in aggregations:
            if isinstance(agg_expr, str):
                # String column name - default to sum()
                from ..expressions.functions import sum as sum_func

                normalized_aggs.append(sum_func(col(agg_expr)))
            elif isinstance(agg_expr, dict):
                # Dictionary syntax: {"column": "function"}
                for col_name, func_name in agg_expr.items():
                    agg_col = self._create_aggregation_from_string(col_name, func_name)
                    normalized_aggs.append(agg_col)
            elif isinstance(agg_expr, Column):
                # Already a Column expression
                normalized_aggs.append(agg_expr)
            else:
                raise ValueError(
                    f"Invalid aggregation type: {type(agg_expr)}. "
                    "Expected Column, str, or Dict[str, str]"
                )

        # For pivoted grouped data, we can only aggregate one column at a time
        # (PySpark behavior - pivot with multiple aggregations requires different syntax)
        if len(normalized_aggs) > 1:
            raise ValueError(
                "Pivoted grouped aggregation supports only one aggregation expression. "
                "Multiple aggregations are not supported with pivot."
            )

        agg_expr = normalized_aggs[0]
        self._validate_aggregation(agg_expr)

        # Extract the value column from the aggregation
        # For sum(col("amount")), we need "amount"
        value_column = self._extract_value_column(agg_expr)

        # Extract the aggregation function name
        agg_func = self._extract_agg_func(agg_expr)

        # If pivot_values is not provided, infer them from the data (PySpark behavior)
        pivot_values = self.pivot_values
        if pivot_values is None:
            # Query distinct values from the pivot column
            distinct_df = DataFrame(plan=self.plan, database=self.parent.database)
            distinct_df = distinct_df.select(col(self.pivot_column)).distinct()
            distinct_rows = distinct_df.collect()
            pivot_values = tuple(
                str(row[self.pivot_column])
                for row in distinct_rows
                if row[self.pivot_column] is not None
            )

            if not pivot_values:
                raise ValueError(
                    f"No distinct values found in pivot column '{self.pivot_column}'. "
                    "Please provide pivot_values explicitly."
                )

        # Create a GroupedPivot logical plan
        plan = operators.grouped_pivot(
            self.plan,
            grouping=self.keys,
            pivot_column=self.pivot_column,
            value_column=value_column,
            agg_func=agg_func,
            pivot_values=pivot_values,
        )
        return DataFrame(plan=plan, database=self.parent.database)

    @staticmethod
    def _extract_value_column(agg_expr: Column) -> str:
        """Extract the column name from an aggregation expression.

        Args:
            agg_expr: Aggregation Column expression (e.g., sum(col("amount")))

        Returns:
            Column name string (e.g., "amount")

        Raises:
            ValueError: If the column cannot be extracted
        """
        if not agg_expr.op.startswith("agg_"):
            raise ValueError("Expected an aggregation expression")

        if not agg_expr.args:
            raise ValueError("Aggregation expression must have arguments")

        # The first argument should be a Column with op="column"
        col_expr = agg_expr.args[0]
        if not isinstance(col_expr, Column):
            raise ValueError("Aggregation must operate on a column")

        if col_expr.op == "column":
            if not col_expr.args:
                raise ValueError("Column expression must have arguments")
            col_name = col_expr.args[0]
            if not isinstance(col_name, str):
                raise ValueError("Column name must be a string")
            return col_name
        else:
            raise ValueError(f"Cannot extract column name from expression: {col_expr.op}")

    @staticmethod
    def _extract_agg_func(agg_expr: Column) -> str:
        """Extract the aggregation function name from an aggregation expression.

        Args:
            agg_expr: Aggregation Column expression (e.g., sum(col("amount")))

        Returns:
            Aggregation function name (e.g., "sum")
        """
        op = agg_expr.op
        if op == "agg_sum":
            return "sum"
        elif op == "agg_avg":
            return "avg"
        elif op == "agg_min":
            return "min"
        elif op == "agg_max":
            return "max"
        elif op == "agg_count" or op == "agg_count_star":
            return "count"
        elif op == "agg_count_distinct":
            return "count_distinct"
        else:
            # Default to sum if unknown
            return "sum"

    @staticmethod
    def _create_aggregation_from_string(column_name: str, func_name: str) -> Column:
        """Create an aggregation Column from a column name and function name string.

        Args:
            column_name: Name of the column to aggregate
            func_name: Name of the aggregation function (e.g., "sum", "avg", "min", "max", "count")

        Returns:
            Column expression for the aggregation

        Raises:
            ValueError: If the function name is not recognized
        """
        from ..expressions.functions import (
            avg,
            count,
            max as max_func,
            min as min_func,
            sum as sum_func,
            count_distinct,
        )

        from typing import Callable

        func_map: Dict[str, Callable[[Column], Column]] = {
            "sum": sum_func,
            "avg": avg,
            "average": avg,  # Alias for avg
            "min": min_func,
            "max": max_func,
            "count": count,
            "count_distinct": count_distinct,
        }

        func_name_lower = func_name.lower()
        if func_name_lower not in func_map:
            raise ValueError(
                f"Unknown aggregation function: {func_name}. "
                f"Supported functions: {', '.join(func_map.keys())}"
            )

        agg_func = func_map[func_name_lower]
        return agg_func(col(column_name))

    @staticmethod
    def _validate_aggregation(expr: Column) -> Column:
        """Validate that an expression is a valid aggregation.

        Args:
            expr: Column expression to validate

        Returns:
            The validated column expression

        Raises:
            ValueError: If the expression is not a valid aggregation
        """
        if not expr.op.startswith("agg_"):
            raise ValueError(
                "Aggregation expressions must be created with moltres aggregate helpers "
                "(e.g., sum(), avg(), count(), min(), max())"
            )
        return expr
